#!/usr/bin/env python3
"""
    Copyright (C) Ηλιάδης Ηλίας, 2019-09-24; Ηλιάδης Ηλίας <iliadis@kekbay.gr>

    This file is part of «test sudoku».

    «test sudoku» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «test sudoku» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «test sudoku».  If not, see <http://www.gnu.org/licenses/>.
"""

"""
        WARNING !!!

Since we are in python, any subclass of this class
will take over any method with the same name.
So do not declare here methods that exist also in the subclass.
(subclass is the class that uses the class here as Base.
Imagine the subclass as a super-duper class,
although in computing superclass is the base class)
"""
#FIXME: correct the version
__version__ = '0.1.28'
VERSIONSTR = 'v. {}'.format(__version__)

try:
    import os
    import sys
    import json

    # Gtk and related
    from gi import require_version as gi_require_version
    gi_require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from gi.repository import Gdk
    from gi.repository import GObject
    # space for extra Gtk imports

    # Configuration and message boxes
    from auxiliary import SectionConfig, OptionConfig
    from auxiliary import MessageBox
    from constants import *

    # Load static methods
    from windowpuzzle_statics import *

    from puzzle import *
    from gamepixbufs import *

except ImportError as eximp:
    print(eximp)
    sys.exit(-1)

class Error(Exception):
    """Base class for exceptions in this module."""
    pass

class WindowPuzzleBase(object):
    #FIXME: fix the docstring.
    """ Main window with all components. """

    def __init__(self, *args, **kwargs):
        # Set the app
        self.myparent = None
        self.custom_args = kwargs['custom_args']

        # bind settings,options to a class variable
        global settings
        settings = self.settings
        global options
        options = self.options

        # Before builder.
        self._run_before_builder()

        # Read GUI from file and retrieve objects from Gtk.Builder
        thebuilder = Gtk.Builder()
        thebuilder.set_translation_domain(self.Application.id)
        try:
            thebuilder.add_from_file(os.path.join(self.Application.custom_args.APP_DIR,
                'ui',
                'windowpuzzle.glade')
                )
        except GObject.GError:
            print("Error reading GUI file")
            raise

        # Fire up the main window
        #We must get the object from builder to set the Application
        #before loading the objects, so that the rest will be loaded
        #using locale
        self.WindowPuzzle = thebuilder.get_object("WindowPuzzle")
        self.WindowPuzzle.set_application(self.Application)
        self._get_from_builder(thebuilder)
        self.mywindow = self.WindowPuzzle
        self._post_initialisations()
        self.mywindow.show()

#********* Auto created "class defs" START ************************************************************
    def _run_before_builder(self):
        self.resize_timer_id = 0
        self.history_difficulty = ""
        #self.continue_last = False

    def _get_from_builder(self, builder):
        """ Create self names for easy access. """
        self.BoxForPuzzle = builder.get_object('BoxForPuzzle')
        self.BoxOnTop = builder.get_object('BoxOnTop')
        self.ButtonDummyTest = builder.get_object('ButtonDummyTest')
        self.ButtonRedo = builder.get_object('ButtonRedo')
        self.ButtonReturn = builder.get_object('ButtonReturn')
        self.ButtonUndo = builder.get_object('ButtonUndo')
        self.DrawingAreaForPuzzle = builder.get_object('DrawingAreaForPuzzle')
        self.DummyLabel = builder.get_object('DummyLabel')
        self.EventBoxForPuzzle = builder.get_object('EventBoxForPuzzle')
        self.ImageClock = builder.get_object('ImageClock')
        self.ImageForButtonRedo = builder.get_object('ImageForButtonRedo')
        self.ImageForButtonUndo = builder.get_object('ImageForButtonUndo')
        self.LabelTimePassed = builder.get_object('LabelTimePassed')
        self.LabelVersion = builder.get_object('LabelVersion')
        self.ListBoxForPieces = builder.get_object('ListBoxForPieces')
        self.MainBox = builder.get_object('MainBox')
        self.WindowPuzzle = builder.get_object('WindowPuzzle')

        # Connect signals existing in the Glade file.
        builder.connect_signals(self)

        # Connect generated by OCPgenerator signals:
        # to builder's main window
        # first are the defaults for window
        self.WindowPuzzle.connect('delete-event', self.on_WindowPuzzle_delete_event)
        self.WindowPuzzle.connect('destroy', self.on_WindowPuzzle_destroy)
        self.WindowPuzzle.connect('size-allocate', self.on_WindowPuzzle_size_allocate)
        self.WindowPuzzle.connect('window-state-event', self.on_WindowPuzzle_window_state_event)
        self.ButtonDummyTest.connect('clicked', self.on_ButtonDummyTest_clicked)
        self.ButtonRedo.connect('clicked', self.on_ButtonRedo_clicked)
        self.ButtonReturn.connect('clicked', self.on_ButtonReturn_clicked)
        self.ButtonUndo.connect('clicked', self.on_ButtonUndo_clicked)

    def _post_initialisations(self):
        """ Do some extra initializations.

        Display the version if a labelVersion is found.
        Set defaults (try to load them from a configuration file):
            - Window size and state (width, height and if maximized)
        Load any custom settings from a configuration file.
        """
        if 'parent' in self.custom_args:#Is a child window, get the parent window
            self.myparent = self.custom_args['parent']
            #if has parent check for transient
            # modality can be false, and parent may not be present
            if 'transient' in self.custom_args:
                self.mywindow.set_transient_for(self.myparent)
                if 'modal' in self.custom_args:
                    self.mywindow.set_modal(True)

        if 'trigger_before_exit' in self.custom_args:
            # must be a function on calling class
            self.trigger_before_exit = self.custom_args['trigger_before_exit']
            self.return_parameters = None

        # Bind message boxes.
        self.MessageBox = MessageBox(self.WindowPuzzle, self.Application)
        self.msg = self.MessageBox.Message
        self.are_you_sure = self.MessageBox.are_you_sure

        # Reset MainWindow to a default or previous size and state.
        width = settings.get('width', 350)
        height = settings.get('height', 350)
        self.DIFFICULTY = options.get("DIFFICULTY", "")
        self.WindowPuzzle.resize(width, height)
        self.WindowPuzzle.set_icon(self.Application.icon)
        if settings.get_bool('maximized', False):
            self.WindowPuzzle.maximize()
        # Set the label for LabelVersion
        self.LabelVersion.set_label("v." + self.Application.custom_args.version)
        self.LabelVersion.set_tooltip_text(_("Version of this window:") + "\n" + VERSIONSTR)

        # Load any other settings here.
        self.pb = None
        self.picker = None
        self.previous_size = 0

        self.init_puzzle_options()
        #check if user wants to continue last game
        if 'continue' in self.custom_args:
            self.last_game = self.custom_args['last_game']
            self.puzzle = APuzzle(last_game = self.last_game,
                    trigger_refresh = self.trigger_from_puzzle,
                    )
            self.WindowPuzzle.set_title(
                self.Application.custom_args.localizedname + \
                " (" + _(self.last_game['difficulty']) + ")")
            self.continue_last = True
        else:
            self.continue_last  = False
            puzzle_string, solution, difficulty = get_new_puzzle(self.DIFFICULTY)
            self.puzzle = APuzzle(puzzle_string = puzzle_string,
                    difficulty = difficulty,
                    trigger_refresh = self.trigger_from_puzzle)
            self.puzzle.solution = solution
            self.WindowPuzzle.set_title(
                self.Application.custom_args.localizedname + \
                " (" + _(difficulty) + ")")

        self.EventBoxForPuzzle.drag_dest_set(0, [],0)
        self.EventBoxForPuzzle.set_events(Gdk.EventMask.ALL_EVENTS_MASK )

#********* Auto created handlers START *********************************
    def on_ButtonDummyTest_clicked(self, widget, *args):
        """ Handler for ButtonDummyTest.clicked. """
        self.create_board()

    def on_ButtonRedo_clicked(self, widget, *args):
        """ Handler for ButtonRedo.clicked. """
        self.puzzle.redo()
        self.create_board(True)

    def on_ButtonReturn_clicked(self, widget, *args):
        """ Handler for ButtonReturn.clicked. """
        self.exit_requested()

    def on_ButtonUndo_clicked(self, widget, *args):
        """ Handler for ButtonUndo.clicked. """
        self.puzzle.undo()
        self.create_board(True)

    def on_DrawingAreaForPuzzle_draw(self, widget, cr, *args):
        """ Handler for DrawingAreaForPuzzle.draw. """
        #TODO: possibly remove time consuming functionality
        if self.board_pixbuf:
            #get actual width and height
            ebw = self.EventBoxForPuzzle.get_allocated_width()
            ebh = self.EventBoxForPuzzle.get_allocated_height()
            board_x = int((ebw - self.board_pixbuf.get_width()) / 2)
            board_y = int((ebh - self.board_pixbuf.get_height()) / 2)
            Gdk.cairo_set_source_pixbuf(cr, self.board_pixbuf, board_x, board_y)
            cr.paint()

    def on_EventBoxForPuzzle_motion_notify_event(self, widget, event, *args):
        """ Handler for EventBoxForPuzzle.motion-notify-event.

        Store the cellnumber to self.mouse_over
        in order to save time in calculations.
        Move the picker window there if hidden.

        Moves the picker over the cell in which the cursor is...
        """
        cell_n , sx, sy = get_real_positions(event.x,
                event.y,
                self.EventBoxForPuzzle,self.board_pixbuf)
        self.mouse_over = cell_n
        if cell_n and self.puzzle.start_cells[cell_n][1] == 0:
            upleftposition = \
                self.EventBoxForPuzzle.translate_coordinates (self.mywindow,
                    sx,
                    sy)
            windowposition = self.mywindow.get_window().get_root_coords(0, 0)
            self.show_picker_at = (upleftposition[0] + windowposition[0] + 1,
                        upleftposition[1] + windowposition[1] + 1)
            if not self.picker.mywindow.is_visible():
                self.picker.mywindow.move(*self.show_picker_at)
        else:
            self.show_picker_at = None

    def on_EventBoxForPuzzle_button_press_event(self, widget, event, *args):
        """ Handler for EventBoxForPuzzle.button-press-event. """
        if self.mouse_over:
            # if picker is visible then hide and move window there
            if self.picker.mywindow.is_visible():
                self.picker.mywindow.hide()
                if self.show_picker_at:
                    self.picker.mywindow.move(*self.show_picker_at)
            else:#picker is hidden. Move there and show
                self.picker.mywindow.move(*self.show_picker_at)
                self.picker.mywindow.show()
                self.clicked_on = self.mouse_over
        else:
            if self.picker.mywindow.is_visible():
                self.picker.mywindow.hide()
        return True

    def on_EventBoxForPuzzle_button_release_event(self, widget, event, *args):
        """ Handler for EventBoxForPuzzle.button-release-event. """
        cell_n, sx, sy = get_real_positions(event.x,
                event.y,
                self.EventBoxForPuzzle, self.board_pixbuf)
        if cell_n:
            if self.puzzle.current_cells[0] != cell_n:
                self.puzzle.current_cells[0] = cell_n
                self.create_board(True)
            return True
        else:
            if self.picker.mywindow.is_visible():
                self.picker.mywindow.hide()

    def on_EventBoxForPuzzle_drag_data_received(self, widget, context, x, y, selection_data, info, time_, *args):
        """ Handler for EventBoxForPuzzle.drag-data-received. """
        cell_n , sx, sy = get_real_positions(x,
                y,
                self.EventBoxForPuzzle,self.board_pixbuf)
        if self.number_picked == CLEAR_AS_NUMBER_PICKED:
            if not self.puzzle.current_cells[cell_n][0]:
                #make it current
                self.puzzle.current_cells[0] = cell_n
                self.set_selected_cell(0)

        elif self.number_picked and (not self.puzzle.current_cells[cell_n][0]):
            #make it current
            self.puzzle.current_cells[0] = cell_n
            self.set_selected_cell(int(self.number_picked))

    def on_EventBoxForPuzzle_drag_drop(self, widget, context, x, y, time_, *args):
        """ Handler for EventBoxForPuzzle.drag-drop. """
        cell_n , sx, sy = get_real_positions(x,
                y,
                self.EventBoxForPuzzle,self.board_pixbuf)
        if cell_n:
            if not self.puzzle.current_cells[cell_n][0]:
                widget.drag_get_data(context, context.list_targets()[-1], time_)

    #TODO: show drop area and change default drag icon
    def on_EventBoxForPuzzle_drag_motion(self, widget, context, x, y, time_, *args):
        """ Handler for EventBoxForPuzzle.drag-motion. """
        cell_n , sx, sy = get_real_positions(x,
                y,
                self.EventBoxForPuzzle,self.board_pixbuf)
        if cell_n != 0:
            if not self.puzzle.current_cells[cell_n][0]:
                Gdk.drag_status(context, Gdk.DragAction.COPY, time_)
                return True
        return False

    def on_EventBoxForPuzzle_focus_in_event(self, widget, event, *args):
        """ Handler for EventBoxForPuzzle.focus-in-event. """
        self.create_board()

    def on_EventBoxForPuzzle_focus_out_event(self, widget, event, *args):
        """ Handler for EventBoxForPuzzle.focus-out-event. """
        self.create_board()

    def on_WindowPuzzle_button_press_event(self, widget, event, *args):
        """ Handler for WindowPuzzle.button-press-event. """
        if self.show_picker_at:
            self.picker.mywindow.move(*self.show_picker_at)
        else:
            if self.picker.mywindow.is_visible():
                self.picker.mywindow.hide()

    def on_WindowPuzzle_configure_event(self, widget, event, *args):
        """ Handler for WindowPuzzle.configure-event. """
        #https://stackoverflow.com/questions/19058392/detecting-when-a-gtk-window-is-done-moving-resizing-by-the-user
        if self.resize_timer_id:
            GObject.source_remove(self.resize_timer_id)
        self.resize_timer_id = GObject.timeout_add(interval=250, function=self.resize_done)
        return False

    def on_WindowPuzzle_key_press_event(self, widget, event, *args):
        """ Handler for WindowPuzzle.key-press-event.

        Prevent list box from focusing on arrow clicks.
        """
        pass

    def on_WindowPuzzle_key_release_event(self, widget, event, *args):
        """ Handler for WindowPuzzle.key-release-event.

        Control combined with "z", "y" and "q" keys
        and Alt combined with "t" key
        are processed here.
        Other Control and Alt combinations are not.

        Some keys set the focus and are processed here.
        These are the arrow keys, backspace and delete keys
        as well as unichar keys used in puzzle.
        Other keys (example tab key) are not processed here.
        """
        txt = Gdk.keyval_name(event.keyval)
        if type(txt) == type(None):
            return False

        unichar = chr(Gdk.keyval_to_unicode(event.keyval))
        txt = txt.replace('KP_', '')
        if event.get_state() & Gdk.ModifierType.CONTROL_MASK:
            if txt.lower() == "z":
                if self.puzzle.history_position > 0:
                    self.puzzle.undo()
                    self.create_board()
                return True
            elif txt.lower() == "y":
                if self.puzzle.history_position < len(self.puzzle.history):
                    self.puzzle.redo()
                    self.create_board()
                return True
            elif txt.lower() == "q":
                self.exit_requested()
                return True
            elif txt.lower() == "h":
                self.exit_requested()
                return True
            return False
        if event.get_state() & Gdk.ModifierType.MOD1_MASK:
            if txt.lower() == "t" : #Alt+t
                self.set_show_timer(not self.SHOW_TIMER)
                return True
            return False
        if txt in ['Up', 'Down', 'Right', 'Left']:
            self.move_selection_by_arrow(txt)
            return True
        if txt in ['BackSpace','Delete']:
            self.EventBoxForPuzzle.grab_focus()
            if not self.puzzle.current_cells[self.puzzle.current_cells[0]][0]:
                self.set_selected_cell(0)
            return True
        if unichar in self.strings_to_use[1:]:
            self.EventBoxForPuzzle.grab_focus()
            if not self.puzzle.current_cells[self.puzzle.current_cells[0]][0]:
                theint = self.strings_to_use.index(unichar)
                self.set_selected_cell(theint)
            return True
        return False

#********* Auto created handlers END ***********************************

#********* Standard handlers START *************************************
    def msg_not_yet(self):
        self.msg(_('Not yet implemented'))

    def on_WindowPuzzle_delete_event(self, widget, event, *args):
        """ Handler for our main window: delete-event. """
        return (self.exit_requested())

    def on_WindowPuzzle_destroy(self, widget, *args):
        """ Handler for our main window: destroy. """
        return (self.exit_requested('from_destroy'))

    def on_WindowPuzzle_size_allocate(self, widget, allocation, *args):
        """ Handler for our main window: size-allocate. """
        self.save_my_size()

    def on_WindowPuzzle_window_state_event(self, widget, event, *args):
        """ Handler for our main window: window-state-event. """
        settings.set('maximized',
            ((int(event.new_window_state) & Gdk.WindowState.ICONIFIED) != Gdk.WindowState.ICONIFIED) and
            ((int(event.new_window_state) & Gdk.WindowState.MAXIMIZED) == Gdk.WindowState.MAXIMIZED)
            )
        self.save_my_size()

#********* Standard handlers END ***************************************
#********* Standard exit defs START *********************************************************
    def exit_requested(self, *args, **kwargs):
        """ Final work before exit. """
        if self.resize_timer_id:
            GObject.source_remove(self.resize_timer_id)
            self.resize_timer_id = 0
        if self.timer_id:
            GObject.source_remove(self.timer_id)
            self.timer_id = 0
        self.WindowPuzzle.set_transient_for()
        self.WindowPuzzle.set_modal(False)
        if self.picker:
            self.picker.mywindow.destroy()
            self.picker = None
        self.set_unhandled_settings()# also saves all settings
        if 'from_destroy' in args:
            return True
        else:
            # Check if we should provide info to caller
            if 'trigger_before_exit' in self.custom_args:
                self.trigger_before_exit(exiting = True,
                    return_parameters = self.return_parameters)
            self.WindowPuzzle.destroy()

    def present(self):
        """ Show the window. """
        pass

    def save_my_size(self):
        """ Save the window size into settings, if not maximized. """
        if not settings.get_bool('maximized', False):
            width, height = self.WindowPuzzle.get_size()
            settings.set('width', width)
            settings.set('height', height)

    def set_unhandled_settings(self):
        """ Set, before exit, any settings not applied during the session.

        Additionally, flush all settings to .conf file.
        Usually options al have been saved.
        """

        # Set any custom settings
        # which where not setted (ex. on some widget's state changed)
        last_puzzle_string = ""
        last_history = ""
        last_history_position = ""
        if not self.puzzle.solved:
            thelists = []
            for a_move in self.puzzle.history:
                thelists.append(":".join([str(x) for x in a_move]))
            last_history = ",".join(thelists)
            last_puzzle_string = self.puzzle.puzzle_string
            last_history_position = self.puzzle.history_position
        options.set('last_puzzle_string', last_puzzle_string)
        options.set('last_history', last_history)
        options.set('last_history_position', last_history_position)
        options.set('last_difficulty', self.puzzle.difficulty)
        options.set('last_seconds_passed', self.seconds_passed())
        options.set('solved', self.puzzle.solved)

        # Save all settings
        settings.save()
#********* Standard exit defs END **************************************
#********* Auto created "class defs" END **************************************************************
    def resize_done(self):
        #https://stackoverflow.com/questions/19058392/detecting-when-a-gtk-window-is-done-moving-resizing-by-the-user
        self.resize_timer_id = 0
        self.check_board_resize()
        return False

    def create_board(self, set_focus=False):
        if set_focus and not self.EventBoxForPuzzle.is_focus():
            self.EventBoxForPuzzle.grab_focus()
        self.board_pixbuf = create_board_pixbuf(self.pb,
                self.puzzle.current_cells,
                self.EventBoxForPuzzle)
        self.DrawingAreaForPuzzle.queue_draw()

    def init_puzzle_options(self):
        self.SHOW_ONLY_BOARD = options.get_bool("show_only_board", False)
        self.SHOW_PIECES = options.get_bool("show_pieces", True)
        self.SHOW_TIMER = options.get_bool("show_timer", True)
        self.font_name = options.get("font_name", "Arial 300")
        font_ratio = options.get("font_ratio","1")
        self.font_ratio = float(font_ratio)

        tmp_strings = options.get("custom_dict",",".join([str(x) for x in THEDICTS['standard'][1:]]))
        custom_dict =[x.strip() for x in tmp_strings.split(",")]
        custom_dict.insert(0, ' ')
        THEDICTS['custom'] = custom_dict
        self.use_dict = options.get("use_dict", 'standard')
        self.strings_to_use = THEDICTS[self.use_dict]

        self.new_colors = {}
        self.new_colors["RGBA_BG_TRANSP"] = Gdk.RGBA(red=1., green=1., blue=1., alpha=0.0)
        self.new_colors["RGBA_BG"] = eval_rgba(
                options.get('RGBA_BG', str(DEFAULT_RGBA['BG'])))
        self.new_colors["RGBA_FG"] = eval_rgba(
                options.get('RGBA_FG', str(DEFAULT_RGBA['FG'])))
        self.new_colors["RGBA_FG_RED"] = eval_rgba(
                options.get('RGBA_FG_RED', str(DEFAULT_RGBA['FG_RED'])))
        self.new_colors["RGBA_BG_CONSTANT"] = eval_rgba(
                options.get('RGBA_BG_CONSTANT', str(DEFAULT_RGBA['BG_CONSTANT'])))
        self.new_colors["RGBA_BG_SELECTED"] = eval_rgba(
                options.get('RGBA_BG_SELECTED', str(DEFAULT_RGBA['BG_SELECTED'])))
        self.new_colors["RGBA_BG_CONSTANT_SELECTED"] = eval_rgba(
                options.get('RGBA_BG_CONSTANT_SELECTED', str(DEFAULT_RGBA['BG_CONSTANT_SELECTED'])))
        self.new_colors["RGBA_PICKER_BG"] = eval_rgba(
                options.get('RGBA_PICKER_BG', str(DEFAULT_RGBA['PICKER_BG'])))
#********* Window class  END***************************************************************************



#!/usr/bin/env python3
"""
    Copyright (C) Ηλίας Ηλιάδης, 2019-09-24; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    This file is part of «Sudoku as puzzle».

    «Sudoku as puzzle» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «Sudoku as puzzle» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «Sudoku as puzzle».  If not, see <http://www.gnu.org/licenses/>.
"""

# No need to check imports.
# They must be already imported by main window class.
import os
import sys
# Gtk and related
from gi import require_version as gi_require_version
gi_require_version('Gtk', '3.0')
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import GdkPixbuf
import cairo

class NumberPicker(object):
    def __init__(self, *args, **kwargs):
        self.passed_args = kwargs['args_to_pass']
        self.pixbufs = self.passed_args['pixbufs']
        self.myparent = self.passed_args['parent']
        self.strings_to_use = self.passed_args['strings_to_use']
        self.response_from_picker = self.passed_args['response_from_picker']

        self.bg_color = self.passed_args['RGBA_PICKER_BG'] #self.pixbufs.newcolors['RGBA_PICKER_BG']

        #dummy initilisation
        self.cell_size = 15

        self.mywindow = Gtk.Window()
        self.mywindow.set_transient_for(self.myparent)
        self.mywindow.set_modal(False)
        self.mywindow.set_decorated(False)
        self.mywindow.set_size_request(1,1)

        self.mywindow.connect('destroy', self.on_Me_destroy)
        self.mywindow.connect('hide', self.on_Me_hide)
        self.mywindow.connect('key-release-event', self.picker_key_release_event)
        self.mywindow.connect('button-press-event', self.on_Me_clicked)

        self.init_for_hide = True

        box = Gtk.Box()
        da = Gtk.DrawingArea()
        da.set_visible(True)
        da.set_hexpand(True)
        da.set_vexpand(True)
        da.connect('draw', self.draw_da)
        da.show()
        box.add(da)
        self.da = da
        box.set_visible(True)
        self.mywindow.add(box)
        box.show()

        self.board_pixbuf = self.create_pixbuf()

    def draw_da(self, widget, cr, *args):
        if self.board_pixbuf:
            box_size = 3 * self.cell_size
            Gdk.cairo_set_source_pixbuf(cr, self.board_pixbuf, 0, 0)
            cr.paint()

    def on_Me_clicked(self, widget, event, *args):
        if event.button == 3:
            self.response_from_picker(number = 0)
        elif event.button == 1:
            x = event.x
            y=event.y
            col, dummy = divmod(x,self.cell_size)
            row, dummy = divmod(y,self.cell_size)
            number = int((3 * row) + col+1)
            self.response_from_picker(number = number)
        return True

    def on_Me_leave_event(self, widget, *args):
        self.response_from_picker(number = -1)
        return True

    def on_Me_hide(self, widget, *args):
        pass

    def on_Me_destroy(self, widget, *args):
        pass

    def picker_key_release_event(self, widget, event, *args):
        """ Handler for mywindow.key-release-event. """
        txt = Gdk.keyval_name(event.keyval)
        if type(txt) == type(None):
            return False
        unichar = chr(Gdk.keyval_to_unicode(event.keyval))
        txt = txt.replace('KP_', '')
        if txt in ['BackSpace','Delete',]:
            self.response_from_picker(number = 0)
            return True
        if txt == 'Escape':
            self.response_from_picker(number = -1)
            return True
        if unichar in self.strings_to_use[1:]:
            theint = self.strings_to_use.index(unichar)
            self.response_from_picker(number = theint)
            return True

    def picker_button_press_event(self, widget, event, *args):
        if event.button == 3:
            self.response_from_picker(number = 0)
        elif event.button == 1:
            number = (3 * args[1]) + (args[0]+1)
            self.response_from_picker(number = number)
        return True

    def create_pixbuf(self):
        one_size = 3 * self.cell_size
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, one_size, one_size)
        cr = cairo.Context(surface)
        w0 = self.cell_size

        number = 0

        for column in range(3):
            for row in range(3):
                number = (3 * row) + (column+1)
                apixbuf = self.pixbufs[number]
                cr.save()
                cr.set_source_rgba(*self.bg_color)
                cr.rectangle(w0*column  ,w0*row , w0, w0)
                cr.fill()
                cr.restore()
                cr.save()
                Gdk.cairo_set_source_pixbuf(cr, apixbuf,w0*column,w0*row)
                cr.paint()
                cr.stroke()
                cr.restore()
        cr.set_line_width(1)
        cr.set_source_rgba(0, 0, 0, 1)
        for acounter in range(3):
            cr.save()
            x1 = y1 = w0 * acounter
            cr.move_to(x1, 0)
            cr.line_to(x1, one_size)
            cr.stroke()
            cr.move_to(0, x1)
            cr.line_to(one_size, x1)
            cr.stroke()
            cr.restore()

        board_pixbuf = Gdk.pixbuf_get_from_surface(surface, 0, 0, one_size, one_size)
        return board_pixbuf

    def change_for_pixbufs(self, pixbufs):
        self.pixbufs = pixbufs
        self.cell_size = 2 + self.pixbufs[1].get_width()
        self.board_pixbuf = self.create_pixbuf()

        self.mywindow.set_size_request(3*self.cell_size - 2 , 3*self.cell_size-2)
        self.mywindow.resize(3*self.cell_size - 2 , 3*self.cell_size -2)




#!/usr/bin/env python3
"""
    Copyright (C) Ηλίας Ηλιάδης, 2019-09-24; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    This file is part of «Sudoku as puzzle».

    «Sudoku as puzzle» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «Sudoku as puzzle» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «Sudoku as puzzle».  If not, see <http://www.gnu.org/licenses/>.
"""

"""
        WARNING !!!

Since we are in python, any subclass of this class
will take over any method with the same name.
So do not declare here methods that exist also in the subclass.
(subclass is the class that uses the class here as Base.
Imagine the subclass as a super-duper class,
although in computing superclass is the base class)
"""
#FIXME: correct the version
__version__ = '0.1.28'
VERSIONSTR = 'v. {}'.format(__version__)

try:
    import os
    import sys

    # Gtk and related
    from gi import require_version as gi_require_version
    gi_require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from gi.repository import Gdk
    from gi.repository import GObject
    # space for extra Gtk imports

    # Configuration and message boxes
    from auxiliary import SectionConfig, OptionConfig
    from auxiliary import MessageBox
    from constants import *

    # Load static methods
    from windowoptions_statics import *

    from puzzle import *
    from gamepixbufs import *

except ImportError as eximp:
    print(eximp)
    sys.exit(-1)

class WindowOptionsBase(object):
    #FIXME: fix the docstring.
    """ Main window with all components. """

    def __init__(self, *args, **kwargs):
        # Set the app
        self.myparent = None
        self.custom_args = kwargs['custom_args']

        # bind settings,options to a class variable
        global settings
        settings = self.settings
        global options
        options = self.options

        # Before builder.
        self._run_before_builder()

        # Read GUI from file and retrieve objects from Gtk.Builder
        thebuilder = Gtk.Builder()
        thebuilder.set_translation_domain(self.Application.id)
        try:
            thebuilder.add_from_file(os.path.join(self.Application.custom_args.APP_DIR,
                'ui',
                'windowoptions.glade')
                )
        except GObject.GError:
            print("Error reading GUI file")
            raise

        # Fire up the main window
        #We must get the object from builder to set the Application
        #before loading the objects, so that the rest will be loaded
        #using locale
        self.WindowOptions = thebuilder.get_object("WindowOptions")
        self.WindowOptions.set_application(self.Application)
        self._get_from_builder(thebuilder)
        self.mywindow = self.WindowOptions
        self._post_initialisations()
        self.mywindow.show()

#********* Auto created "class defs" START ************************************************************
    def _run_before_builder(self):
        self.board_pixbuf = None
        #self.new_colors["RGBA_BG_TRANSP"] = Gdk.RGBA(red=1., green=1., blue=1., alpha=0.0)
        self.dicts_widgets = {}
        self.custom_chars_widgets = {}
        THEDICTS["custom"] = THEDICTS["standard"].copy()

    def _get_from_builder(self, builder):
        """ Create self names for easy access. """
        self.AdjustmentForRatio = builder.get_object('AdjustmentForRatio')
        self.BoxForBottom = builder.get_object('BoxForBottom')
        self.BoxForDifficulty = builder.get_object('BoxForDifficulty')
        self.BoxForDrawingArea = builder.get_object('BoxForDrawingArea')
        self.BoxForLegend = builder.get_object('BoxForLegend')
        self.BoxMain = builder.get_object('BoxMain')
        self.BoxOnButtonCheck = builder.get_object('BoxOnButtonCheck')
        self.BoxOnTop = builder.get_object('BoxOnTop')
        self.ButtonBgColorConstant = builder.get_object('ButtonBgColorConstant')
        self.ButtonBgColorConstantSelected = builder.get_object('ButtonBgColorConstantSelected')
        self.ButtonBgColorEmpty = builder.get_object('ButtonBgColorEmpty')
        self.ButtonBgColorSelected = builder.get_object('ButtonBgColorSelected')
        self.ButtonDefaults = builder.get_object('ButtonDefaults')
        self.ButtonExit = builder.get_object('ButtonExit')
        self.ButtonFgColorRed = builder.get_object('ButtonFgColorRed')
        self.ButtonFontName = builder.get_object('ButtonFontName')
        self.ButtonForeColor = builder.get_object('ButtonForeColor')
        self.ButtonPickerBgColor = builder.get_object('ButtonPickerBgColor')
        self.ButtonRefresh = builder.get_object('ButtonRefresh')
        self.ButtonSave = builder.get_object('ButtonSave')
        self.CheckOnlyBoard = builder.get_object('CheckOnlyBoard')
        self.CheckShowRemaining = builder.get_object('CheckShowRemaining')
        self.CheckShowTimer = builder.get_object('CheckShowTimer')
        self.DrawingAreaForDemo = builder.get_object('DrawingAreaForDemo')
        self.DummyLabelColors = builder.get_object('DummyLabelColors')
        self.DummyLabelFont = builder.get_object('DummyLabelFont')
        self.Eb1_1 = builder.get_object('Eb1_1')
        self.Eb1_2 = builder.get_object('Eb1_2')
        self.Eb1_3 = builder.get_object('Eb1_3')
        self.Eb1_4 = builder.get_object('Eb1_4')
        self.Eb1_5 = builder.get_object('Eb1_5')
        self.Eb1_6 = builder.get_object('Eb1_6')
        self.Eb1_7 = builder.get_object('Eb1_7')
        self.Eb1_8 = builder.get_object('Eb1_8')
        self.Eb1_9 = builder.get_object('Eb1_9')
        self.GridForDictionaries = builder.get_object('GridForDictionaries')
        self.GridForFontAndColors = builder.get_object('GridForFontAndColors')
        self.GridForLegend = builder.get_object('GridForLegend')
        self.GridPresentOptions = builder.get_object('GridPresentOptions')
        self.ImageOnButtonRefresh = builder.get_object('ImageOnButtonRefresh')
        self.LabelAppearance = builder.get_object('LabelAppearance')
        self.LabelDifficulty = builder.get_object('LabelDifficulty')
        self.LabelLegend = builder.get_object('LabelLegend')
        self.LabelLegend1 = builder.get_object('LabelLegend1')
        self.LabelLegend2 = builder.get_object('LabelLegend2')
        self.LabelLegend3 = builder.get_object('LabelLegend3')
        self.LabelLegend4 = builder.get_object('LabelLegend4')
        self.LabelLegend5 = builder.get_object('LabelLegend5')
        self.LabelLegend6 = builder.get_object('LabelLegend6')
        self.LabelLegend7 = builder.get_object('LabelLegend7')
        self.LabelLegend8 = builder.get_object('LabelLegend8')
        self.LabelLegend9 = builder.get_object('LabelLegend9')
        self.LabelLegendDescription = builder.get_object('LabelLegendDescription')
        self.LabelLetter1_1 = builder.get_object('LabelLetter1_1')
        self.LabelLetter1_2 = builder.get_object('LabelLetter1_2')
        self.LabelLetter1_3 = builder.get_object('LabelLetter1_3')
        self.LabelLetter1_4 = builder.get_object('LabelLetter1_4')
        self.LabelLetter1_5 = builder.get_object('LabelLetter1_5')
        self.LabelLetter1_6 = builder.get_object('LabelLetter1_6')
        self.LabelLetter1_7 = builder.get_object('LabelLetter1_7')
        self.LabelLetter1_8 = builder.get_object('LabelLetter1_8')
        self.LabelLetter1_9 = builder.get_object('LabelLetter1_9')
        self.LabelNumberingSystem = builder.get_object('LabelNumberingSystem')
        self.LabelOnButtonRefresh = builder.get_object('LabelOnButtonRefresh')
        self.RadioButtonAny = builder.get_object('RadioButtonAny')
        self.RadioButtonForLetters = builder.get_object('RadioButtonForLetters')
        self.RadioButtoneasy = builder.get_object('RadioButtoneasy')
        self.RadioButtonexpert = builder.get_object('RadioButtonexpert')
        self.RadioButtonintermediate = builder.get_object('RadioButtonintermediate')
        self.RadioButtonsimple = builder.get_object('RadioButtonsimple')
        self.ScaleForRatio = builder.get_object('ScaleForRatio')
        self.WindowOptions = builder.get_object('WindowOptions')

        # Connect signals existing in the Glade file.
        builder.connect_signals(self)

        # Connect generated by OCPgenerator signals:
        # to builder's main window
        # first are the defaults for window
        self.WindowOptions.connect('delete-event', self.on_WindowOptions_delete_event)
        self.WindowOptions.connect('destroy', self.on_WindowOptions_destroy)
        self.WindowOptions.connect('size-allocate', self.on_WindowOptions_size_allocate)
        self.WindowOptions.connect('window-state-event', self.on_WindowOptions_window_state_event)
        self.ButtonDefaults.connect('clicked', self.on_ButtonDefaults_clicked)
        self.ButtonExit.connect('clicked', self.on_ButtonExit_clicked)
        self.ButtonRefresh.connect('clicked', self.on_ButtonRefresh_clicked)
        self.ButtonSave.connect('clicked', self.on_ButtonSave_clicked)
        self.CheckOnlyBoard.connect('toggled', self.on_CheckOnlyBoard_toggled)
        self.CheckShowRemaining.connect('toggled', self.on_CheckShowRemaining_toggled)
        self.CheckShowTimer.connect('toggled', self.on_CheckShowTimer_toggled)

    def _post_initialisations(self):
        """ Do some extra initializations.

        Display the version if a labelVersion is found.
        Set defaults (try to load them from a configuration file):
            - Window size and state (width, height and if maximized)
        Load any custom settings from a configuration file.
        """
        if 'parent' in self.custom_args:#Is a child window, get the parent window
            self.myparent = self.custom_args['parent']
            #if has parent check for transient
            # modality can be false, and parent may not be present
            if 'transient' in self.custom_args:
                self.mywindow.set_transient_for(self.myparent)
                if 'modal' in self.custom_args:
                    self.mywindow.set_modal(True)

        if 'trigger_before_exit' in self.custom_args:
            # must be a function on calling class
            self.trigger_before_exit = self.custom_args['trigger_before_exit']
            self.return_parameters = None

        # Bind message boxes.
        self.MessageBox = MessageBox(self.WindowOptions, self.Application)
        self.msg = self.MessageBox.Message
        self.are_you_sure = self.MessageBox.are_you_sure

        # Reset MainWindow to a default or previous size and state.
        width = settings.get('width', 350)
        height = settings.get('height', 350)
        self.WindowOptions.set_title(_("Options") + " - " + self.Application.custom_args.localizedname)
        self.WindowOptions.resize(width, height)
        self.WindowOptions.set_icon(self.Application.icon)
        if settings.get_bool('maximized', False):
            self.WindowOptions.maximize()

        # Load any other settings here.
        self.pb = None
        self.picker = None
        self.previous_size = 0

        self.init_defaults()
        self.GridForLegend.override_background_color(Gtk.StateFlags.NORMAL, DEFAULT_RGBA['BG'])

#********* Auto created handlers START *********************************
    def on_ButtonColor_color_set(self, widget, *args):
        """ Handler for any ButtonColor.color-set. """
        rbga = widget.get_rgba()
        if widget == self.ButtonBgColorEmpty:
            self.all_colors['RGBA_BG'] = rbga
        elif widget == self.ButtonForeColor:
            self.all_colors['RGBA_FG'] = rbga
        elif widget == self.ButtonFgColorRed:
            self.all_colors['RGBA_FG_RED'] = rbga
        elif widget == self.ButtonBgColorConstant:
            self.all_colors['RGBA_BG_CONSTANT'] = rbga
        elif widget == self.ButtonBgColorSelected:
            self.all_colors['RGBA_BG_SELECTED'] = rbga
        elif widget == self.ButtonBgColorConstantSelected:
            self.all_colors['RGBA_BG_CONSTANT_SELECTED'] = rbga
        elif widget == self.ButtonPickerBgColor:
            self.all_colors['RGBA_PICKER_BG'] = rbga

    def on_ButtonDefaults_clicked(self, widget, *args):
        """ Handler for ButtonDefaults.clicked. """
        self.set_defaults()

    def on_ButtonExit_clicked(self, widget, *args):
        """ Handler for ButtonExit.clicked. """
        self.exit_requested()
        return True

    def on_ButtonFontName_font_set(self, widget, *args):
        """ Handler for ButtonFontName.font-set. """
        self.font_name = widget.get_font_name()
        self.fix_font_size()

    def on_ButtonRefresh_clicked(self, widget, *args):
        """ Handler for ButtonRefresh.clicked. """
        self.create_board()
        return True

    def on_ButtonSave_clicked(self, widget, *args):
        """ Handler for ButtonSave.clicked. """
        self.save_new_options()
        self.exit_requested()
        return True

    def on_CheckOnlyBoard_toggled(self, widget, *args):
        """ Handler for CheckOnlyBoard.toggled. """
        self.SHOW_ONLY_BOARD = self.CheckOnlyBoard.get_active()
        self.CheckShowRemaining.set_sensitive(not self.SHOW_ONLY_BOARD)
        self.CheckShowTimer.set_sensitive(not self.SHOW_ONLY_BOARD)

    def on_CheckShowRemaining_toggled(self, widget, *args):
        """ Handler for CheckShowRemaining.toggled. """
        self.SHOW_PIECES = self.CheckShowRemaining.get_active()

    def on_CheckShowTimer_toggled(self, widget, *args):
        """ Handler for CheckShowTimer.toggled. """
        self.SHOW_TIMER = self.CheckShowTimer.get_active()

    def on_DrawingAreaForDemo_draw(self, widget, cr, *args):
        """ Handler for DrawingAreaForDemo.draw. """
        if self.board_pixbuf:
            cr.save()
            Gdk.cairo_set_source_pixbuf(cr, self.board_pixbuf, 0, 0)
            cr.paint()
            cr.restore()

    def on_RadioButtonDifficulty_toggled(self, widget, *args):
        """ Handler for any RadioButton for difficulty toggled. """
        if widget.get_active():
            self.DIFFICULTY = widget.get_name()

    def on_ScaleForRatio_change_value(self, widget, scroll, new_value, *args):
        """ Handler for ScaleForRatio.change-value. """
        if new_value <= 1 and new_value >=0.4:
            self.font_ratio = new_value

    def on_WindowOptions_key_release_event(self, widget, event, *args):
        """ Handler for WindowOptions.key-release-event. """
        txt = Gdk.keyval_name(event.keyval)
        if type(txt) == type(None):
            return False

        unichar = chr(Gdk.keyval_to_unicode(event.keyval))
        txt = txt.replace('KP_', '')
        if event.get_state() & Gdk.ModifierType.CONTROL_MASK:
            if txt.lower() == "s":
                self.save_new_options()
                self.exit_requested()
                return True
            elif txt.lower() == "p":
                self.create_board()
                return True
            elif txt.lower() == "h":
                self.exit_requested()
                return True
            elif txt.lower() == "q":
                self.exit_requested()
                return True
            elif txt.lower() == "a":
                self.MessageBox.AboutBox()
                return True
#********* Auto created handlers END ***********************************

#********* Standard handlers START *************************************
    def msg_not_yet(self):
        """ Standarized message for not implemented functionality. """
        self.msg(_('Not yet implemented'))

    def on_WindowOptions_delete_event(self, widget, event, *args):
        """ Handler for our main window: delete-event. """
        return (self.exit_requested())

    def on_WindowOptions_destroy(self, widget, *args):
        """ Handler for our main window: destroy. """
        return (self.exit_requested('from_destroy'))

    def on_WindowOptions_size_allocate(self, widget, allocation, *args):
        """ Handler for our main window: size-allocate. """
        self.save_my_size()

    def on_WindowOptions_window_state_event(self, widget, event, *args):
        """ Handler for our main window: window-state-event. """
        settings.set('maximized',
            ((int(event.new_window_state) & Gdk.WindowState.ICONIFIED) != Gdk.WindowState.ICONIFIED) and
            ((int(event.new_window_state) & Gdk.WindowState.MAXIMIZED) == Gdk.WindowState.MAXIMIZED)
            )
        self.save_my_size()

#********* Standard handlers END ***************************************
#********* Standard exit defs START *********************************************************
    def exit_requested(self, *args, **kwargs):
        """ Final work before exit. """
        self.WindowOptions.set_transient_for()
        self.WindowOptions.set_modal(False)
        self.set_unhandled_settings()# also saves all settings
        if 'from_destroy' in args:
            return True
        else:
            # Check if we should provide info to caller
            if 'trigger_before_exit' in self.custom_args:
                self.trigger_before_exit(exiting = True,
                    return_parameters = self.return_parameters)
            self.WindowOptions.destroy()

    def present(self):
        """ Show the window. """
        pass

    def save_my_size(self):
        """ Save the window size into settings, if not maximized. """
        if not settings.get_bool('maximized', False):
            width, height = self.WindowOptions.get_size()
            settings.set('width', width)
            settings.set('height', height)

    def set_unhandled_settings(self):
        """ Set, before exit, any settings not applied during the session.

        Additionally, flush all settings to .conf file.
        """
        # Set any custom settings
        # which where not setted (ex. on some widget's state changed)
        # Save all settings
        settings.save()
#********* Standard exit defs END **************************************
#********* Auto created "class defs" END **************************************************************

    def fill_dicts(self):
        thegrid = self.GridForDictionaries
        self.RadioButtonForLetters.tag = 'standard'
        self.dicts_widgets['standard'] = self.RadioButtonForLetters
        self.RadioButtonForLetters.connect('toggled', self.on_RadioButtonDict_toggled, self.dicts_widgets['standard'])
        for rowcounter in range(len(DICTSLIST))[1:-1]:
            dict_name = DICTSLIST[rowcounter]
            thedict = THEDICTS[dict_name]
            theoption = Gtk.RadioButton(_(dict_name).title())
            theoption.join_group(self.RadioButtonForLetters)
            theoption.connect('toggled', self.on_RadioButtonDict_toggled, theoption)
            theoption.tag = dict_name
            thegrid.attach(theoption,0,rowcounter,1,1)
            self.dicts_widgets[dict_name] = theoption
            for x in range(9):
                label = Gtk.Label(thedict[x+1])
                thegrid.attach(label,x+1,rowcounter,1,1)
        rowcounter = len(DICTSLIST)-1
        dict_name = DICTSLIST[rowcounter]
        thedict = THEDICTS[dict_name]
        theoption = Gtk.RadioButton(_(dict_name).title())
        theoption.join_group(self.RadioButtonForLetters)
        theoption.connect('toggled', self.on_RadioButtonDict_toggled, theoption)
        theoption.tag = dict_name
        thegrid.attach(theoption,0,rowcounter,1,1)
        for lettercounter in range(9):
            entry = Gtk.Entry()
            entry.set_text(thedict[lettercounter+1])
            entry.set_width_chars(2)#fixing the stupid GtkEnonEntryButEntryWithGarbageandmore
            entry.set_has_frame (False)
            entry.set_max_width_chars(1)
            entry.entry_tag = lettercounter+1
            entry.connect('changed', self.on_any_custom_char_changed, lettercounter+1)
            entry.connect('button-release-event', self.on_any_custom_char_clicked)
            entry.set_sensitive(False)
            thegrid.attach(entry,lettercounter+1,rowcounter,1,1)
            self.custom_chars_widgets[lettercounter] = entry
        self.dicts_widgets[dict_name] = theoption
        thegrid.show_all()
        #self.dicts_widgets[self.use_dict].set_active(True)

    def on_RadioButtonDict_toggled(self, widget, *args):
        #self.use_dict = args[0].tag
        if widget.get_active():
            self.use_dict = widget.tag
            is_the_custom = (widget.tag == 'custom')
            for anentry in self.custom_chars_widgets:
                self.custom_chars_widgets[anentry].set_sensitive(is_the_custom)

    def on_any_custom_char_changed(self, widget, *args):
        THEDICTS['custom'][args[0]] = widget.get_text()

    def on_any_custom_char_clicked(self, widget, *args):
        widget.select_region(0,-1)

    def init_defaults(self):
        self.DIFFICULTY = 'easy' #simple, easy, intermediate, expert
        self.SHOW_ONLY_BOARD = False
        self.SHOW_PIECES = True
        self.SHOW_TIMER = True
        self.font_name = "Arial 400"
        self.font_ratio = float(0.95)
        self.use_dict = "standard"
        self.all_colors = {}#ALL_COLORS.copy()
        self.all_colors['RGBA_BG'] = DEFAULT_RGBA['BG']
        self.all_colors['RGBA_FG'] = DEFAULT_RGBA['FG']
        self.all_colors['RGBA_FG_RED'] = DEFAULT_RGBA['FG_RED']
        self.all_colors['RGBA_BG_CONSTANT'] = DEFAULT_RGBA['BG_CONSTANT']
        self.all_colors['RGBA_BG_SELECTED'] = DEFAULT_RGBA['BG_SELECTED']
        self.all_colors['RGBA_BG_CONSTANT_SELECTED'] = DEFAULT_RGBA['BG_CONSTANT_SELECTED']
        self.all_colors["RGBA_BG_TRANSP"] = DEFAULT_RGBA['BG_TRANSP']
        self.all_colors["RGBA_PICKER_BG"] = DEFAULT_RGBA['PICKER_BG']

    def set_defaults(self):
        self.init_defaults()
        self.show_current_options()

    def load_options(self):
        """ Load options.

        First will set the defaults.
        Then will load them from a user configuration file.
        """
        self.init_defaults()
        self.DIFFICULTY = options.get("DIFFICULTY", self.DIFFICULTY)
        if self.DIFFICULTY not in ["simple", "easy", "intermediate", "expert", "any"]:
            self.DIFFICULTY = 'easy'
        self.SHOW_ONLY_BOARD = options.get_bool("show_only_board", self.SHOW_ONLY_BOARD)
        self.SHOW_PIECES = options.get_bool("show_pieces", self.SHOW_PIECES)
        self.SHOW_TIMER = options.get_bool("show_timer", self.SHOW_TIMER)

        self.font_name = options.get("font_name", self.font_name)
        self.fix_font_size()
        self.ButtonFontName.set_font_name(self.font_name)
        font_ratio = options.get("font_ratio",str(self.font_ratio))
        self.font_ratio = float(font_ratio)
        strings = options.get("custom_dict","")
        strings_list = [x.strip() for x in strings.split(",")]
        if len(strings_list) == 9:
            strings_list.insert(0, ' ')
            THEDICTS['custom'] = strings_list
        self.use_dict = options.get("use_dict", self.use_dict)
        self.all_colors["RGBA_BG"] = eval_rgba(
                options.get('RGBA_BG', str(self.all_colors['RGBA_BG'])))
        self.all_colors["RGBA_FG"] = eval_rgba(
                options.get('RGBA_FG', str(self.all_colors['RGBA_FG'])))
        self.all_colors["RGBA_FG_RED"] = eval_rgba(
                options.get('RGBA_FG_RED', str(self.all_colors['RGBA_FG_RED'])))
        self.all_colors["RGBA_BG_CONSTANT"] = eval_rgba(
                options.get('RGBA_BG_CONSTANT', str(self.all_colors['RGBA_BG_CONSTANT'])))
        self.all_colors["RGBA_BG_SELECTED"] = eval_rgba(
                options.get('RGBA_BG_SELECTED', str(self.all_colors['RGBA_BG_SELECTED'])))
        self.all_colors["RGBA_BG_CONSTANT_SELECTED"] = eval_rgba(
                options.get('RGBA_BG_CONSTANT_SELECTED', str(self.all_colors['RGBA_BG_CONSTANT_SELECTED'])))
        self.all_colors["RGBA_PICKER_BG"] = eval_rgba(
                options.get('RGBA_PICKER_BG', str(self.all_colors['RGBA_PICKER_BG'])))
        self.show_current_options()

    def show_current_options(self):
        """ Show active options.

        Forces update of any changes.
        Calls set_rgba_in_buttons for updating the colors
        in buttons.
        """
        self.CheckOnlyBoard.set_active(self.SHOW_ONLY_BOARD)
        self.CheckShowRemaining.set_active(self.SHOW_PIECES)
        self.CheckShowTimer.set_active(self.SHOW_TIMER)
        self.ScaleForRatio.set_value(self.font_ratio)
        self.set_rgba_in_buttons()
        if not (self.use_dict in self.dicts_widgets):
            #show a delayed message
            GObject.timeout_add(interval=50, function=self.stupid_timer_1)
            self.use_dict = "standard"
        self.dicts_widgets[self.use_dict].set_active(True)
        for anumber in range(9):
            self.custom_chars_widgets[anumber].set_text(THEDICTS["custom"][anumber+1])
        self.create_board()
        for anwidget in self.BoxForDifficulty.get_children():
            if anwidget.get_name() == self.DIFFICULTY:
                anwidget.set_active(True)
                break

    def stupid_timer_1(self):
        self.msg(_("Dictionary {0} not found!\nUsing standard.").format(self.use_dict,),
            boxtype = 'ERROR')
        return False

    def set_rgba_in_buttons(self):
        self.ButtonBgColorEmpty.set_rgba(self.all_colors['RGBA_BG'])
        self.ButtonForeColor.set_rgba(self.all_colors['RGBA_FG'])
        self.ButtonFgColorRed.set_rgba(self.all_colors['RGBA_FG_RED'])
        self.ButtonBgColorConstant.set_rgba(self.all_colors['RGBA_BG_CONSTANT'])
        self.ButtonBgColorSelected.set_rgba(self.all_colors['RGBA_BG_SELECTED'])
        self.ButtonBgColorConstantSelected.set_rgba(self.all_colors['RGBA_BG_CONSTANT_SELECTED'])
        self.ButtonPickerBgColor.set_rgba(self.all_colors['RGBA_PICKER_BG'])

    def save_new_options(self):
        try:
            options.set("show_only_board", self.SHOW_ONLY_BOARD)
            options.set("show_pieces", self.SHOW_PIECES)
            options.set("show_timer", self.SHOW_TIMER)
            options.set("DIFFICULTY", self.DIFFICULTY)
            options.set("font_name", self.font_name)
            options.set("font_ratio",str(self.font_ratio))
            options.set("custom_dict",','.join([str(x) for x in THEDICTS["custom"][1:]]))
            options.set("use_dict", self.use_dict)
            options.set("RGBA_BG", str(self.all_colors['RGBA_BG']))
            options.set("RGBA_FG", str(self.all_colors['RGBA_FG']))
            options.set("RGBA_FG_RED", str(self.all_colors['RGBA_FG_RED']))
            options.set("RGBA_BG_CONSTANT", str(self.all_colors['RGBA_BG_CONSTANT']))
            options.set("RGBA_BG_SELECTED", str(self.all_colors['RGBA_BG_SELECTED']))
            options.set("RGBA_BG_CONSTANT_SELECTED", str(self.all_colors['RGBA_BG_CONSTANT_SELECTED']))
            options.set("RGBA_PICKER_BG", str(self.all_colors['RGBA_PICKER_BG']))
            self.msg(_("Options saved!"), boxtype = 'INFO')
            return
        except Exception as e:
            print(e)
        self.msg(_("Options NOT saved!"), boxtype = 'ERROR')

    def create_board(self):
        self.pb = GamePixBufs(self.mywindow, THEDICTS[self.use_dict], self.font_name, self.all_colors, self.font_ratio)
        self.pb.fit_pixbufs(OPTIONS_BOARD_SIZE // 3)
        self.board_pixbuf = create_board_pixbuf_for_options(self.pb)
        self.DrawingAreaForDemo.queue_draw()

    def fix_font_size(self):
        self.font_name = self.font_name.split(" ",-1)[0] + " 10"

#********* Window class  END***************************************************************************



#!/usr/bin/env python3
"""
    Copyright (C) Ηλίας Ηλιάδης, 2019-09-24; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    This file is part of «Sudoku as puzzle».

    «Sudoku as puzzle» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «Sudoku as puzzle» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «Sudoku as puzzle».  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys

WHERE_AM_I = os.path.dirname(os.path.abspath(os.path.realpath(__file__)))
#mypackagefolder = os.path.abspath(os.path.join(WHERE_AM_I,".."))
sys.path.insert(0, WHERE_AM_I)

#FIXME: change package name
from myapplication import MyApplication

def main(*args, **kwargs):
    START_DIR = os.path.dirname(os.path.abspath('.'))
    myapp = MyApplication(START_DIR=START_DIR)
    myapp.run(**kwargs)

if __name__ == '__main__':
    """ Main entry for the entire application. """
    main(sys.argv)



#!/usr/bin/env python3
"""
    Copyright (C) Ηλίας Ηλιάδης, 2019-09-24; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    This file is part of «Sudoku as puzzle».

    «Sudoku as puzzle» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «Sudoku as puzzle» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «Sudoku as puzzle».  If not, see <http://www.gnu.org/licenses/>.
"""

from copy import deepcopy

ROWS = []
COLS = []
SQUARES = []
for xcounter in range(3):
    col = xcounter * 3
    colindex = col * 9
    rowindex = col
    for ycounter in range(3):
        ROWS.append( [1+ colindex + (ycounter*9) + x for x in range(9)])
        COLS.append( [1+ rowindex + ycounter + 9 * x for x in range(9)])
        row = ycounter * 3
        square = col * 9 + row
        SQUARES.append( [square+1, square+2, square+3,
                10+square, 11+square, 12+square,
                19+square, 20+square, 21+square])

class APuzzle():
    def __init__(self,  puzzle_string=None,
                #thelist=None, history_as_string=None,
                #history_position=None,
                trigger_refresh=None,
                difficulty=None,
                last_game = None):
        self.solved = False

        self.numbers_shown = {}
        self.all_pieces = {}
        if trigger_refresh:
            self.trigger_refresh = trigger_refresh
        else:
            raise "Cannot continue." + \
            " No method for trigger_refresh provided."
            sys.exit(-1)
        for acounter in range(9):
            self.all_pieces[acounter+1] = 9
        self.all_pieces[0] = 81
        self.remaining = {}
        self.history = []
        self.history_position = 0
        self.solution = []
        self.difficulty = difficulty
        if last_game:
            self.puzzle_string = last_game['puzzle_string']
            self.history_position = last_game['history_position']
            self.history_string = last_game['history_string']
            # create last history
            self.poke_history(last_game['history_string'])
            self.difficulty = last_game['difficulty']

        else:
            self.puzzle_string = puzzle_string
            self.start_cells = self.init_cells()
            #keep a separate list for running game
            self.current_cells = deepcopy(self.start_cells)
        self.set_remaining()

    def dummy_method(self, *args):
        pass

    def init_cells(self):
        dummy = []
        for acounter in range(81):
            s = self.puzzle_string[acounter]
            is_constant = True if s != "." else False
            theint = int(s) if s != "." else 0
            dummy.append ([is_constant, theint, False])
        dummy.insert(0, 1)
        return dummy

    def poke_history(self, history_as_string):
        self.start_cells = self.init_cells()
        #keep a separate list for running game
        self.current_cells = deepcopy(self.start_cells)
        for amovestring in self.history_string.strip().split(","):
            themove = amovestring.split(":")
            new_cell = int(themove[0])
            previous_value = int(themove[1])
            new_value = int(themove[2])
            seconds_passed = int(themove[3])
            self.history.append([new_cell,
                    previous_value,
                    new_value,
                    seconds_passed])

        #play up to history position
        for acounter in range(self.history_position):
            amove = self.history[acounter]
            the_cell = amove[0]
            the_int = amove[2]
            self.current_cells[the_cell][1] = the_int

    def undo(self):
        _last = self.history[self.history_position-1]
        thecell = _last[0]
        theint = _last[1]
        self.current_cells[thecell][1] = theint
        self.current_cells[0] = thecell
        self.history_position -= 1
        self.check_puzzle()

    def redo(self):
        _next = self.history[self.history_position]
        thecell = _next[0]
        theint = _next[2]
        self.current_cells[thecell][1] = theint
        self.current_cells[0] = thecell
        self.history_position += 1
        self.check_puzzle()

    def set_cell(self, thecell, theint, secondspassed):
        """ Insert a number in a cell and append action to history. """
        previous = self.current_cells[thecell][1]
        self.current_cells[thecell][1] = theint
        if self.history_position < len(self.history):
            #remove rest of history
            self.history = self.history[:self.history_position]
        self.history.append([thecell, previous, theint, secondspassed])
        self.history_position = len(self.history)
        self.check_puzzle()

    def check_puzzle(self):
        """ Check if there are duplicates and if puzzle is solved. """

        self.set_remaining()
        self.dups = []
        for alist in ROWS:
            self.enlist_duplicates(alist)
        for alist in COLS:
            self.enlist_duplicates(alist)
        for alist in SQUARES:
            self.enlist_duplicates(alist)
        all_filled = True
        for acounter in range(81):
            #set "duplicate-ness" of any duplicate cell
            self.current_cells[acounter+1][2] = (acounter+1 in self.dups)
        #self.set_remaining()
        self.solved = ((self.remaining[0] == 0) and (len(self.dups) == 0 ))
        self.trigger_refresh()

    def enlist_duplicates(self, thelist):
        """ Create a list of cell numbers that are duplicate in a 9 array. """
        cell_nums = [self.current_cells[x][1] for x in thelist]
        for acounter in range(8):
            if cell_nums[acounter] and cell_nums[acounter] in cell_nums[acounter+1:]:
                adup1pos = acounter
                adup2pos = acounter + 1 + cell_nums[acounter+1:].index(cell_nums[acounter])
                adup1cell = thelist[adup1pos]
                adup2cell = thelist[adup2pos]
                if adup1cell not in self.dups:
                    self.dups.append(adup1cell)
                if adup2cell not in self.dups:
                    self.dups.append(adup2cell)

    def set_remaining(self):
        """ Fill the list of numbers not already poked in puzzle board. """
        self.remaining = self.all_pieces.copy()
        for acounter in range(1, 82):
            theint = self.current_cells[acounter][1]
            if theint:
                self.remaining[theint] -= 1
                self.remaining[0] -= 1





#!/usr/bin/env python3
"""
    Copyright (C) Ηλίας Ηλιάδης, 2019-09-24; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    This file is part of «Sudoku as puzzle».

    «Sudoku as puzzle» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «Sudoku as puzzle» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «Sudoku as puzzle».  If not, see <http://www.gnu.org/licenses/>.
"""

"""
        WARNING !!!

Since we are in python, any method here will override base class.
So do not declare here methods that exist also in the base class.
"""
#FIXME: correct the version
__version__ = '0.1.28'
VERSIONSTR = 'v. {}'.format(__version__)

try:
    import os
    import sys

    # Gtk and related
    from gi import require_version as gi_require_version
    gi_require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from gi.repository import Gdk
    from gi.repository import GObject

    # Configuration and message boxes
    from auxiliary import SectionConfig, OptionConfig
    from auxiliary import MessageBox
    from constants import *

    # Load base window class and static methods
    from windowoptions_base import WindowOptionsBase
    from windowoptions_statics import *

    from picker import *
    from gamepixbufs import *

except ImportError as eximp:
    print(eximp)
    sys.exit(-1)

class WindowOptionsClass(WindowOptionsBase):
    #FIXME: fix the docstring.
    """ Main window with all components. """

    def __init__(self, application, *args, **kwargs):
        self.Application = application
        # define and bind settings, options to a class variable
        global settings
        global options
        settings = SectionConfig(self.Application.id, self.__class__.__name__)
        options = OptionConfig(self.Application.id)
        self.settings = settings
        self.options = options

        #Place here any initilizations before loading the ui
        #Passed custom_args will be evaluated while loading the ui

        super().__init__(self,  *args, **kwargs)

        #place here initilisations after loading the ui
        #first create the widgets
        self.fill_dicts()
        #then load current options
        self.load_options()
        self.show_current_options()






#!/usr/bin/env python3
"""
    Auxiliary classes.

    Copyright (C) Ηλίας Ηλιάδης, 2018; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    Auxiliary classes are free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Auxiliary classes are distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Auxiliary classes.  If not, see <http://www.gnu.org/licenses/>.
"""

#FIXME: correct the version
__version__ = '0.0.20'
VERSIONSTR = 'v. {}'.format(__version__)

ERROR_IMPORT_LIBRARIES_FAIL = -1

#Name of folder inside .config folder
CONFIG_BASE_DIR = "OCP"

try:
    import os
    import pathlib

    from gi import require_version as gi_require_version
    gi_require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from configparser import ConfigParser
except ImportError as eximp:
    print(eximp)
    sys.exit(ERROR_IMPORT_LIBRARIES_FAIL)

settings = None # Keep window related options
options = None #Keep application wide options in a 'General Options' section

class MessageBox():
    def __init__(self, parent_window, application):
        self.parent = parent_window
        self.app = application
        self.INFO_DIR = os.path.join(self.app.custom_args.APP_DIR, 'info')
        self.show = self.Message

    def AboutBox(self):
        aboutdialog = Gtk.AboutDialog()
        #show the localized name together with the international
        aboutdialog.set_program_name(self.app.custom_args.localizedname)
        aboutdialog.set_version("«" + self.app.custom_args.name + "»" + ' v.' + self.app.custom_args.version)
        AUTHORSFILE = os.path.join(self.INFO_DIR, 'AUTHORS')
        ARTISTSFILE = os.path.join(self.INFO_DIR, 'ARTISTS')
        COPYRIGHTFILE = os.path.join(self.INFO_DIR, 'COPYRIGHT')
        COMMENTSFILE = os.path.join(self.INFO_DIR, 'COMMENTS')
        TRANSLATORSFILE = os.path.join(self.INFO_DIR, 'TRANSLATORS')
        if os.path.exists(AUTHORSFILE):
            with open(AUTHORSFILE, mode='rt', encoding='utf-8') as f:
                aboutdialog.set_authors(f.readlines())
        if os.path.exists(ARTISTSFILE):
            with open(ARTISTSFILE, mode='rt', encoding='utf-8') as f:
                aboutdialog.set_artists(f.readlines())
        if os.path.exists(COPYRIGHTFILE):
            with open(COPYRIGHTFILE, mode='rt', encoding='utf-8') as f:
                aboutdialog.set_copyright(f.read())
        if os.path.exists(COMMENTSFILE):
            with open(COMMENTSFILE, mode='rt', encoding='utf-8') as f:
                aboutdialog.set_comments(f.read())
        if os.path.exists(TRANSLATORSFILE):
            with open(TRANSLATORSFILE, mode='rt', encoding='utf-8') as f:
                aboutdialog.set_translator_credits(f.read())
        aboutdialog.set_transient_for(self.parent)
        aboutdialog.set_logo(self.app.icon)
        aboutdialog.run()
        aboutdialog.destroy()

    def Message(self, message, buttons='', boxtype = 'INFO', use_as_title = None, *args):
        """Simple MessageBox.

        Keyword arguments:
        buttons -- string containing one or more of (delimited or no):
            YES, OK, CANCEL, NO
            defaults to: OK
        boxtype -- a string declaring the type of dialog. One of:
            INFO, ERROR, WARNING, QUESTION
            defaults to: INFO
        usethetitle -- override title with this string
        """
        Buttons = tuple()
        if 'YES' in buttons:
            Buttons += (_('_Yes'), Gtk.ResponseType.OK)
        elif 'OK' in buttons:
            Buttons += (_('_OK'), Gtk.ResponseType.OK)
        if 'NO' in buttons:
            Buttons += (_('_No'), Gtk.ResponseType.CANCEL)
        elif 'CANCEL' in buttons:
            Buttons += (_('_Cancel'), Gtk.ResponseType.CANCEL)
        if len(Buttons) == 0:
            Buttons += (_('OK'), Gtk.ResponseType.OK)
        if boxtype == 'ERROR':
            MessageType = Gtk.MessageType.ERROR
            title = _('ERROR')
        elif boxtype == 'WARNING':
            MessageType = Gtk.MessageType.WARNING
            title = _('WARNING')
        elif boxtype == 'QUESTION':
            MessageType = Gtk.MessageType.QUESTION
            title = _('QUESTION')
        else:
            MessageType = Gtk.MessageType.INFO
            title = _('INFO')
        if use_as_title:
            title = use_as_title
        #dialog = Gtk.MessageDialog(self.parent, 0, MessageType, Buttons, title)
        dialog = Gtk.MessageDialog(parent = self.parent, type=MessageType, buttons = Buttons, text = title)
        dialog.format_secondary_text(message)
        dialog.set_title(self.app.custom_args.localizedname)
        dialog.vbox.set_spacing (3)
        for a in dialog.vbox:
            for b in a:
                if type(b) == Gtk.ButtonBox:
                    b.set_halign(Gtk.Align.CENTER)
        response = dialog.run()
        dialog.destroy()
        return True if response == Gtk.ResponseType.OK else False

    def are_you_sure(self, addmessage):
        message = _('You are about to:\n') + addmessage + '\n\n'
        message += _('Are you sure?')
        return self.Message(message, buttons='YESCANCEL', boxtype = 'WARNING')

class SectionConfig():
    """A very simplified configuration editor for multi usage.

    Read and write only one section from a conf file in user's home directory.
    Can be called from multiple "places"
    using the same filename, but different section names.
    Config file is only opened and closed at init and at save.
    All modification during the existance of the instance
    are done in the instance's dir only for the specified section.

    """
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.parser:
            self.save()

    def __init__(self, base_name, section):
        """Init the python's configparser.

        Read existing keys for "my" section only, if any.
        Modify only "my" section.
        Readings are done using python's configparser idioms.
        Methods:
        - get
        - set
        - get_bool
        - save

        """

        self.S = section

        self.ok = False
        self.parser = None
        self.configfile = os.path.join(os.path.expanduser('~'), '.config', CONFIG_BASE_DIR, base_name + '.conf')
        create_empty_config(self.configfile,"SectionConfig")
        self.parser = ConfigParser()
        self.parser.add_section(self.S)
        self.parser_error = True
        #ensure path up to file exists
        #os.makedirs(os.path.dirname(self.configfile), exist_ok=True)
        #read the conf to a local parser and close it
        thetext = self.get_configfile_content()
        if self.parser_error:#an error occured. file unusable
            return
        if thetext != '':
            #get only the contents of self.S section
            dummyparser = ConfigParser()
            dummyparser.read_string(thetext)
            if dummyparser.has_section(self.S):
                for name, value in dummyparser.items(self.S):
                    self.parser.set(self.S, name, value)

    def get_configfile_content(self):
        try:
            with open(self.configfile, 'r') as f:
                thetext = f.read()
                self.parser_error = False
                return thetext
        except FileNotFoundError:
            pass
        except Exception:
            return None
        #the file not found
        try:
            with open(self.configfile, 'w') as f:
                self.parser_error = False
                return ''
        except Exception:
            return None

    def get(self, thekey, thedefault):
        """Get a value for a key based on the type of the provided default value.

        Return a default value if key does not exists.
        """
        if self.parser.has_option(self.S, thekey):
            if type(thedefault) == str:
                return self.parser.get(self.S, thekey)
            elif type(thedefault) == bool:
                return self.parser.getboolean(self.S, thekey)
            else:
                theoldkey = None
                try:#first if int
                    theoldkey = self.parser.getint(self.S, thekey)
                except ValueError:#otherwise a str
                    theoldkey = self.parser.get(self.S, thekey)
                except Exception as ex:#other error
                    print('LocalConfig other exception', repr(ex))
                    return theoldkey
                return theoldkey
        else:
            return thedefault

    def get_bool(self, thekey, thedefault):
        """Get a boolean value for a key. """
        b = self.get(thekey, thedefault)
        if type(b) == bool:
            return b
        if type(b) == str:
            if b.isalpha:
                return (b == 'False')
            else:
                return ( 0 == int(b))

    def set(self, thekey, thevalue):
        """Set the value for a key.

        Always convert to str.
        """
        try:
            self.parser.set(self.S, thekey, str(thevalue))
        except:
            print('(DEBUGprint) Option NOT setted (section, option,value)', self.S, thekey, str(thevalue))

    def save(self):
        """Save "my" section to the disk.

        1. Read all sections.
        2. Replace "my" section.
        3. Write all to disk (may re-arragne sections).
        """
        try:
            #open (a.k.a. and lock) the config file
            with open(self.configfile, 'r+') as f:
                #read the content
                thetext = f.read()
                #create a dummy parser with the content
                dummyparser = ConfigParser()
                dummyparser.read_string(thetext)
                if dummyparser.has_section(self.S):
                    #remove all keys for my section
                    dummyparser.remove_section(self.S)
                dummyparser.add_section(self.S)
                #if I have anything to write
                if self.parser.has_section(self.S):
                    #append the keys
                    #print('len(self.parser[self.S])',len(self.parser[self.S]))
                    for name, value in self.parser.items(self.S):
                        dummyparser.set(self.S, name, value)
                #write and free the conf file
                f.seek(0)
                f.truncate()
                dummyparser.write(f)
        except Exception as ex:
            print('write exception ', repr(ex),'\nwhile writing settings file: ', self.configfile)

class OptionConfig():
    """A very simplified configuration editor for multi usage.

    Use class SectionConfig to read and write
    section 'General Options' from a conf file
    in user's home directory.
    All modifications are dumped immediately.

    """
    def __init__(self, base_name):
        """ Initialize the class.

        Parameters
        -------
        base_name : string
            the name of the file (without path)
        """
        self.base_name = base_name
        self.general_section = 'General Options'

    def get(self, thekey, thedefault):
        with SectionConfig(self.base_name, self.general_section) as cf:
            return cf.get(thekey, thedefault)

    def get_bool(self, thekey, thedefault):
        with SectionConfig(self.base_name, self.general_section) as cf:
            return cf.get_bool(thekey, thedefault)

    def set(self, thekey, thevalue):
        with SectionConfig(self.base_name, self.general_section) as cf:
            cf.set(thekey, thevalue)

def create_empty_config(fullpath, test):
    thepath, base_name = os.path.split(fullpath)
    if not os.path.exists(thepath):
        try:
            pathlib.Path(thepath).mkdir(parents=True, exist_ok=True)
            pathlib.Path(fullpath).touch(exist_ok=True)
            #print("Successfully created config.")
        except Exception as e:
            print('Unable to open for writing or create %s (either folders or the file).' % fullpath, e)
            sys.exit(-3)



#!/usr/bin/env python3
"""
    Copyright (C) Ηλίας Ηλιάδης, 2019-09-24; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    This file is part of «Sudoku as puzzle».

    «Sudoku as puzzle» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «Sudoku as puzzle» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «Sudoku as puzzle».  If not, see <http://www.gnu.org/licenses/>.
"""

"""
        WARNING !!!

Since we are in python, any method here will override base class.
So do not declare here methods that exist also in the base class.
"""
#FIXME: correct the version
__version__ = '0.1.28'
VERSIONSTR = 'v. {}'.format(__version__)

try:
    import os
    import sys
    import datetime
    import random

    # Gtk and related
    from gi import require_version as gi_require_version
    gi_require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from gi.repository import Gdk
    from gi.repository import GObject

    # Configuration and message boxes
    from auxiliary import SectionConfig, OptionConfig
    from auxiliary import MessageBox
    from constants import *

    # Load base window class and static methods
    from windowpuzzle_base import WindowPuzzleBase
    from windowpuzzle_statics import *

    from picker import *
    from gamepixbufs import *

except ImportError as eximp:
    print(eximp)
    sys.exit(-1)

class WindowPuzzleClass(WindowPuzzleBase):
    #FIXME: fix the docstring.
    """ Main window with all components. """

    def __init__(self, application, *args, **kwargs):
        self.Application = application
        # define and bind settings, options to a class variable
        global settings
        global options
        settings = SectionConfig(self.Application.id, self.__class__.__name__)
        options = OptionConfig(self.Application.id)
        self.settings = settings
        self.options = options

        #Place here any initilizations before loading the ui
        #Passed custom_args will be evaluated while loading the ui
        self.targets = None # Gtk.TargetList.new([])
        self.number_picked = None
        self.remaining_ebs = {}
        self.show_picker_at = None

        super().__init__(self,  *args, **kwargs)

        #place here initilisations after loading the ui
        self.one_size = 0
        self.mouse_over = 0

        self.pb = GamePixBufs(self.mywindow, self.strings_to_use,
                self.font_name, self.new_colors,
                font_ratio = self.font_ratio)

        if self.SHOW_ONLY_BOARD:
            self.SHOW_PIECES = False
            self.SHOW_TIMER = False
            self.BoxOnTop.set_visible(False)
        if self.SHOW_PIECES:
            self.init_remaining()
        else:
            self.ListBoxForPieces.set_visible(False)
        self.set_show_timer(self.SHOW_TIMER)

        self.check_board_resize()
        self._create_picker()

        self.trigger_from_puzzle()
        self.EventBoxForPuzzle.grab_focus()
        self.start_time = datetime.datetime.utcnow()
        if 'continue' in self.custom_args:
            self.start_time = datetime.datetime.utcnow() - \
                    datetime.timedelta(seconds = int(self.last_game['last_seconds_passed']))

        self.timer_id = GObject.timeout_add(interval=500, function=self.show_new_time)

    def _create_picker(self):
        """ Create the picker window.

        Must be called only once!"""
        some_custom_args = {}
        some_custom_args['response_from_picker'] = self.response_from_picker
        some_custom_args['parent'] = self.mywindow
        some_custom_args['pixbufs'] = self.pb.picker_pixbufs
        some_custom_args['strings_to_use'] = self.strings_to_use
        some_custom_args['RGBA_PICKER_BG'] = self.new_colors['RGBA_PICKER_BG']
        self.picker = NumberPicker(args_to_pass=some_custom_args)
        self.picker.mywindow.hide()

    def on_remaining_pressed(self, widget, event, *args):
        """ Handler for any "remaining pieces label" button-press-event.

        Triggered if label with remaining pieces is pressed. """
        if self.puzzle.remaining[args[0]]>0:
            self.remaining_Image.set_from_pixbuf(self.pb.fitted_pixbufs[args[0]])
        else:
            self.remaining_Image.set_from_pixbuf(CLEAR_ICON)

    def _clear_icon_drag_begin(self, widget, context):
        self.number_picked = CLEAR_AS_NUMBER_PICKED
        Gtk.drag_set_icon_pixbuf(context, CLEAR_ICON, 0, CLEAR_ICON.get_height())

    def _clear_icon_drag_end(self, widget, context):
        self.number_picked = None

    def _eb_drag_begin(self, widget, context):
        #set here the drag icon, this is a connect_after event
        #will override the default dragicon
        if len(widget.get_children()[0].get_label().strip()):
            self.number_picked = widget.my_tag
            Gtk.drag_set_icon_pixbuf(context,
                    self.pb.picker_pixbufs[int(self.number_picked)],
                    self.pb.picker_pixbufs[int(self.number_picked)].get_width(),
                    self.pb.picker_pixbufs[int(self.number_picked)].get_height())
        else:
            self.number_picked = None

    def _eb_drag_end(self, widget, context):
        self.number_picked = None
        self.remaining_Image.set_from_pixbuf(CLEAR_ICON)

    def init_remaining(self):
        self.labels_remaining = {}
        self.ebs_remaining = {}
        thelistbox = self.ListBoxForPieces
        allchildren = thelistbox.get_children()
        for achild in allchildren[:]:
            achild.destroy()
        labeltext = _('Remaining numbers:')
        label = Gtk.Label(labeltext)
        #prevent resize of right listbox
        if len(labeltext) < 12:
            label.set_label(labeltext + " " * 2 * (12-len(labeltext)))
        label.set_visible(True)
        label.set_can_focus (False)
        thelistbox.add(label)
        for number_counter in range(9):
            number_to_show = self.strings_to_use[number_counter+1]
            remaining = 9
            label = Gtk.Label(number_to_show * remaining)
            label.set_xalign(0)
            label.set_can_focus (False)
            eb = Gtk.EventBox()
            eb.add(label)
            self.labels_remaining[number_counter+1] = label
            eb.connect('button-press-event', self.on_remaining_pressed, number_counter+1)
            label.set_tooltip_text(PLACE_NUMBER_TEXT.format(number_to_show))
            thelistbox.add(eb)
            eb.set_visible(True)
            eb.my_tag = str(number_counter+1)
            label.set_visible(True)
            eb_data = {'widget':eb,
                'handler1':None,
                'handler2':None}
            self.remaining_ebs[number_counter] = eb_data
        label = Gtk.Label("81")
        label.set_xalign (0.5)
        label.set_can_focus (False)
        self.LabelRemaining = label
        eb = Gtk.EventBox()
        eb.add(label)
        thelistbox.insert(eb,1)
        eb.set_visible(True)
        label.set_visible(True)
        theimage = Gtk.Image.new_from_pixbuf(CLEAR_ICON)
        theimage.set_visible(True)
        self.remaining_Image = theimage
        eb = Gtk.EventBox()
        eb.add(theimage)
        thelistbox.add(eb)
        eb.drag_source_set(Gdk.ModifierType.BUTTON1_MASK,
            self.targets,
            Gdk.DragAction.COPY)
        eb.connect_after("drag-begin", self._clear_icon_drag_begin)
        eb.connect("drag-end", self._clear_icon_drag_end)
        eb.drag_source_add_text_targets()
        self.ListBoxForPieces.show_all()
        self.show_remaining()

    def response_from_picker(self, *args, **kwargs):
        self.picker.mywindow.hide()
        if 'thenumbertupple' not in kwargs:
            if 'number' not in kwargs:
                return
        if self.clicked_on:
            self.clicked_on = None
            if kwargs['number'] == -1:return
            self.set_selected_cell(kwargs['number'])

    def check_board_resize(self):
        one_size = min(self.EventBoxForPuzzle.get_allocated_height(), self.EventBoxForPuzzle.get_allocated_width())
        if one_size != self.previous_size:
            ameasure = 3 * int(int(one_size/9)/3)
            self.pb.fit_pixbufs(ameasure)
            self.board_pixbuf = create_board_pixbuf(self.pb,
                    self.puzzle.current_cells,
                    self.EventBoxForPuzzle)
            self.previous_size = self.board_pixbuf.get_width()
            #print("check_board_resize", self.picker)
            if self.picker:
                self.picker.change_for_pixbufs(self.pb.picker_pixbufs)
            self.DrawingAreaForPuzzle.queue_draw()

    def show_new_time(self):
        secondspassed = self.seconds_passed()
        minutes, seconds = divmod(secondspassed, 60)
        if minutes>59:
            hours, minutes = divmod(minutes, 60)
            thestr = '{:02}:{:02}:{:02}'.format(hours, minutes, seconds)
        else:
            thestr = '{:02}:{:02}'.format( minutes, seconds)
        if self.SHOW_TIMER:
            self.LabelTimePassed.set_label(thestr)
        return True

    def move_selection_by_arrow(self, arrowtxt):
        if not self.EventBoxForPuzzle.is_focus():
            self.EventBoxForPuzzle.grab_focus()
        row, column = divmod(self.puzzle.current_cells[0]-1, 9)
        if arrowtxt == "Up":
            row = (row-1) if (row>0) else 8
        elif arrowtxt == "Down":
            row = (row+1) if (row<8) else 0
        elif arrowtxt == "Left":
            column = (column-1) if (column>0) else 8
        elif arrowtxt == "Right":
            column = (column+1) if (column<8) else 0
        self.puzzle.current_cells[0] = (9 * row) + (column+1)
        self.create_board()
        return True

    def set_selected_cell(self, anint):
        if not self.EventBoxForPuzzle.is_focus():
            self.EventBoxForPuzzle.grab_focus()
        # do not append to history if same
        if self.puzzle.current_cells[self.puzzle.current_cells[0]][1] != \
                    anint:
            self.puzzle.set_cell(self.puzzle.current_cells[0], anint, self.seconds_passed())
        self.create_board()
        if self.puzzle.solved:
            if self.timer_id:
                GObject.source_remove(self.timer_id)
                self.timer_id = 0
            ok = self.msg(_("Bravo. Puzzle solved."))
            self.exit_requested()
        return True

    def seconds_passed(self):
        """ Return real time seconds passed."""
        new_time = datetime.datetime.utcnow()
        return int( (new_time - self.start_time).total_seconds())

    def show_remaining(self):
        if not self.SHOW_PIECES:return
        self.LabelRemaining.set_label( str(self.puzzle.remaining[0]))
        for acounter in range(9):
            label = self.labels_remaining[acounter+1]
            thestr1 = self.strings_to_use[acounter+1]
            if self.puzzle.remaining[acounter+1]:
                self.labels_remaining[acounter+1].set_label(
                        self.puzzle.remaining[acounter+1] *
                        self.strings_to_use[acounter+1])
            else:
                self.labels_remaining[acounter+1].set_label("")
            #if there are remaining
            no_remainings = (self.puzzle.remaining[acounter+1] < 1)
            thehandler = self.remaining_ebs[acounter]['handler1']
            if no_remainings:
                self.labels_remaining[acounter+1].set_tooltip_text(ALL_PLACED_TEXT.format(thestr1))
                if thehandler: self.unset_the_drag_for(acounter)
            else:
                self.labels_remaining[acounter+1].set_tooltip_text(PLACE_NUMBER_TEXT.format(thestr1))
                if thehandler == None: self.set_the_drag_for(acounter)

    def set_the_drag_for(self, acounter):
        """ Set the EventBox for «acounter» as drag source.

        Caller must check if not already set
        by examining if «handler1» is not None."""
        self.remaining_ebs[acounter]['widget'].drag_source_set(Gdk.ModifierType.BUTTON1_MASK,
            self.targets,
            Gdk.DragAction.COPY)
        eb_h = self.remaining_ebs[acounter]['widget'].connect_after("drag-begin", self._eb_drag_begin)
        self.remaining_ebs[acounter]['handler1'] = eb_h
        eb_h = self.remaining_ebs[acounter]['widget'].connect("drag-end", self._eb_drag_end)
        self.remaining_ebs[acounter]['handler2'] = eb_h
        self.remaining_ebs[acounter]['widget'].drag_source_add_text_targets()

    def unset_the_drag_for(self, acounter):
        """ Unset the EventBox for «acounter» as drag source.

        Caller must check if not already has been unset
        by examining if «handler1» is None."""
        self.remaining_ebs[acounter]['widget'].drag_source_unset()
        self.remaining_ebs[acounter]['widget'].disconnect(self.remaining_ebs[acounter]['handler1'])
        self.remaining_ebs[acounter]['handler1'] = None
        self.remaining_ebs[acounter]['widget'].disconnect(self.remaining_ebs[acounter]['handler2'])
        self.remaining_ebs[acounter]['handler2'] = None

    def show_undo_redo(self):
        self.ButtonUndo.set_sensitive((len(self.puzzle.history) > 0) and (self.puzzle.history_position > 0))
        self.ButtonRedo.set_sensitive(self.puzzle.history_position < len(self.puzzle.history))

    def set_show_timer(self, avalue):
        """ Hide or show the label for time based on preferences."""
        if self.SHOW_ONLY_BOARD:
            self.SHOW_TIMER  = False
        else:
            self.SHOW_TIMER = avalue
        if self.SHOW_TIMER:
            self.LabelTimePassed.show()
            self.ImageClock.show()
        else:
            self.LabelTimePassed.hide()
            self.ImageClock.hide()

    def trigger_from_puzzle(self, *args):
        """ Show remaining pieces and set the sensitivity
        of undo and redo buttons.

        Triggered when a move is done in the puzzle class."""
        self.show_remaining()
        self.show_undo_redo()




#!/usr/bin/env python3
"""
    Copyright (C) Ηλίας Ηλιάδης, 2019-09-24; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    This file is part of «Sudoku as puzzle».

    «Sudoku as puzzle» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «Sudoku as puzzle» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «Sudoku as puzzle».  If not, see <http://www.gnu.org/licenses/>.
"""

try:
    import os
    import re
    # Localization
    import locale
    import gettext

    # Gtk and related
    from gi import require_version as gi_require_version
    gi_require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from gi.repository import Gdk
    from gi.repository import GObject
    from gi.repository import GdkPixbuf
    # space for extra Gtk imports

except ImportError as eximp:
    print(eximp)
    sys.exit(-1)

#avoiding unsecure eval
RGBA_re = 'Gdk.RGBA\(red=(?P<RED>.*?), green=(?P<GREEN>.*?), blue=(?P<BLUE>.*?), alpha=(?P<ALPHA>.*?)\)'

def eval_rgba(rgba_str):
    """ Securely avaluate a Gdk.RGBA string. """
    amatch = re.match(RGBA_re, rgba_str.strip())
    if amatch:
        return eval(rgba_str.strip())
        # #print(amatch.groupdict())
        # thedict = amatch.groupdict()
        return Gdk.RGBA(red = float(thedict['RED']),
            green=float(thedict['GREEN']),
            blue=float(thedict['BLUE']),
            alpha=float(thedict['ALPHA']))
    return None

def RGBAARRAY2int(an_RGBA_AS_ARRAY):
    return int('%02x%02x%02x%02x' % (
                int(an_RGBA_AS_ARRAY[0]*255),
                int(an_RGBA_AS_ARRAY[1]*255),
                int(an_RGBA_AS_ARRAY[2]*255),
                int(an_RGBA_AS_ARRAY[3]*255)),
                16)
def rgba_int(an_rgba):
    return int('%02x%02x%02x%02x' % (
                int(an_rgba.red*255),
                int(an_rgba.green*255),
                int(an_rgba.blue*255),
                int(an_rgba.alpha*255)),
                16)

CLEAR_AS_NUMBER_PICKED = -7

DEFAULT_RGBA = {}
DEFAULT_RGBA['BG'] = Gdk.RGBA(red=1., green=1., blue=1., alpha=1.0)
DEFAULT_RGBA['FG'] = Gdk.RGBA(red=0., green=0., blue=0., alpha=1.0)
DEFAULT_RGBA['FG_RED'] = Gdk.RGBA(red=1., green=0., blue=0., alpha=1.0)
DEFAULT_RGBA['BG_CONSTANT'] = Gdk.RGBA(red=0.85, green=0.85, blue=0.85, alpha=1.0)
DEFAULT_RGBA['BG_SELECTED'] = Gdk.RGBA(red=0.6, green=0.8, blue=0.9, alpha=1.0)
DEFAULT_RGBA['BG_CONSTANT_SELECTED'] = Gdk.RGBA(red=0.6, green=0.7, blue=0.9, alpha=1.0)
DEFAULT_RGBA['PICKER_BG'] = Gdk.RGBA(red=0., green=1., blue=0., alpha=1.0)
DEFAULT_RGBA['BG_TRANSP'] = Gdk.RGBA(red=1., green=1., blue=1., alpha=0.0)

OPTIONS_BOARD_SIZE = 150

TMPDEFAULTPUZZLE = '83....5....29..6....7..8.39..8....2..4.....569............3...8.83.2..6.7.5...29.'

#Create localized constant strings
#WARNING!!! This must be the same as in the APP main
APPID = "org.kekbay.sudokuaspuzzleocp"
APPNAME = "Sudoku as puzzle"
WHERE_AM_I = os.path.dirname(os.path.abspath(os.path.realpath(__file__)))
LOCALE_DIR = os.path.join(WHERE_AM_I, 'localedir')

locale.setlocale(locale.LC_ALL, '')
locale.bindtextdomain(APPID, LOCALE_DIR)
gettext.install(APPID, LOCALE_DIR)

CLEAR_ICON = GdkPixbuf.Pixbuf.new_from_file(os.path.abspath(os.path.join(WHERE_AM_I, "icons","edit-clear.png")))
CLEAR_ICON_NO_DROP = GdkPixbuf.Pixbuf.new_from_file(os.path.abspath(os.path.join(WHERE_AM_I, "icons","edit-delete.png")))

#make them available to translators
dummy = _('standard')
dummy = _('arabic')
dummy = _('greek')
dummy = _('chinese')
dummy = _('persian')
dummy = _('tibetan')
dummy = _('custom')

DICTSLIST = ['standard', 'arabic', 'greek', 'chinese',
        'persian', 'tibetan', 'custom']
THEDICTS = {}
THEDICTS['standard'] = [' ','1','2','3','4','5','6','7','8','9']
THEDICTS['arabic'] = [' ','١','٢','٣','٤','٥','٦','٧','٨','٩']
THEDICTS['greek'] = [' ','α','β','γ','δ','ε','ς','ζ','η','θ']
THEDICTS['chinese'] = [' ','一','二','三','四','五','六','七','八','九']
THEDICTS['persian'] = [' ','۱','۲','۳','۴','۵','۶','۷','۸','۹']
THEDICTS['tibetan'] = [ ' ','༡','༢','༣','༤','༥','༦','༧','༨','༩']
THEDICTS['custom'] = THEDICTS['standard'].copy()


ALL_PLACED_TEXT = _("You have placed all «{}».")
PLACE_NUMBER_TEXT = _("Drag «{}» to the board...")



#!/usr/bin/env python3
"""
    Copyright (C) Ηλίας Ηλιάδης, 2019-09-24; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    This file is part of «Sudoku as puzzle».

    «Sudoku as puzzle» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «Sudoku as puzzle» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «Sudoku as puzzle».  If not, see <http://www.gnu.org/licenses/>.
"""

"""
        WARNING !!!

Since we are in python, any method here will override base class.
So do not declare here methods that exist also in the base class.
"""
#FIXME: correct the version
__version__ = '0.1.28'
VERSIONSTR = 'v. {}'.format(__version__)


TESTMARGIN = 2
try:
    import os
    import sys
    import math
    from copy import deepcopy
    import subprocess

    # Gtk and related
    from gi import require_version as gi_require_version
    gi_require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from gi.repository import Gdk, GdkPixbuf, Pango
    import cairo
    gi_require_version('PangoCairo', '1.0')
    from gi.repository import PangoCairo

    from constants import *

except ImportError as eximp:
    print(eximp)
    sys.exit(-1)

def create_board_pixbuf(pb, game_cells, dummyda):
    if pb == None: return

    one_size = min(dummyda.get_allocated_width(), dummyda.get_allocated_height())
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, one_size, one_size)
    cr = cairo.Context(surface)
    x0 = y0 = int((one_size-2) // 9)
    real_size = TESTMARGIN + 9 * x0
    selected = game_cells[0]
    if not dummyda.has_focus():
        selected = None

    for column in range(9):
        for row in range(9):
            number = (9 * row) + (column+1)
            shownum = game_cells[number][1]
            isconstant = game_cells[number][0]
            is_red = game_cells[number][2]
            cr.save() #save 1
            if (number == selected) and isconstant:
                cr.set_source_rgba(*pb.newcolors["RGBA_BG_CONSTANT_SELECTED"])
            elif (number == selected):
                cr.set_source_rgba(*pb.newcolors["RGBA_BG_SELECTED"])
            elif isconstant:
                cr.set_source_rgba(*pb.newcolors["RGBA_BG_CONSTANT"])
            else:
                cr.set_source_rgba(*pb.newcolors["RGBA_BG"])
            if is_red:
                thelist = pb.fitted_pixbufs_R
            else:
                thelist = pb.fitted_pixbufs
            if shownum:
                # first a square with selected color
                cr.save()#save 2
                cr.rectangle(x0*column,x0*row, x0, x0)
                cr.fill()
                cr.restore()
                #then print a scaled num (red or black)
                cr.save()
                apixbuf = thelist[shownum]
                Gdk.cairo_set_source_pixbuf(cr, apixbuf,x0*column,x0*row)
                cr.paint()
                cr.stroke()
                cr.restore() #save 2
                cr.restore() #save 1
            else:
                #an empty
                cr.save()
                cr.rectangle(x0*column,x0*row, x0, x0)
                cr.fill()
                cr.restore() #save 1
    cr.set_line_width(1)
    cr.set_source_rgba(0, 0, 0, 1)
    for acounter in range(9)[1:]:
        cr.save()
        if 0 == (acounter % 3) :
            cr.set_line_width(3)
        else:
            cr.set_line_width(1)
        x1 = y1 = x0 * acounter
        cr.move_to(x1, 0)
        cr.line_to(x1, real_size)
        cr.stroke()
        cr.move_to(0, x1)
        cr.line_to(real_size, x1)
        cr.stroke()
        cr.restore()
    w = h = real_size #- 4
    cr.set_line_width(3)
    cr.set_source_rgba(0, 0, 0, 1)
    cr.rectangle(1, 1, w-TESTMARGIN, h-TESTMARGIN)
    cr.stroke()
    board_pixbuf = Gdk.pixbuf_get_from_surface(surface, 0, 0, real_size, real_size)
    return board_pixbuf

# TODO: get it from qqwing's dynamic library.
def get_new_puzzle(difficulty = "easy"):
    """ Get a puzzle from qqwing.

    Return:
        - the puzzle as a string
        - the solution as a string
        - the difficulty (to show in case of "any")
    """
    while True:
        # print(difficulty)
        new_puzzle_output = subprocess.run(['qqwing', "--generate", "--difficulty",
                difficulty, "--one-line", "--solution",
                "--stats", "--count-solutions",
                "--symmetry", "random"], stdout=subprocess.PIPE).stdout.decode('utf-8')
        # print(new_puzzle_output)
        if "unique" in new_puzzle_output:
            new_puzzle_lines = new_puzzle_output.splitlines()
            puzzle_start = new_puzzle_lines[0]
            puzzle_solution = new_puzzle_lines[1]
            difficulty = new_puzzle_lines[-1].split()[-1].lower()
            # return b.split('\n')[0]
            return puzzle_start, puzzle_solution, difficulty

def get_real_positions(x, y, eb, pixbuf):
    cell_number = 0
    eb_w = eb.get_allocated_width()
    eb_h = eb.get_allocated_height()
    pb_one_size = pixbuf.get_width()
    pb_size_2 = pixbuf.get_height()

    start_x = int((eb_w - pb_one_size) / 2 )
    start_y = int((eb_h - pb_one_size) / 2 )
    cell_size = int(pb_one_size / 9)

    if x < start_x or x > start_x + pb_one_size-TESTMARGIN or y < start_y or y > start_y + pb_one_size -TESTMARGIN:
        return cell_number, start_x, start_y

    col = int((x - start_x) / cell_size)
    row = int((y - start_y) / cell_size)
    cell_number = int(row) * 9 + int(col+1)
    picker_x = start_x + col * cell_size
    picker_y = start_y + row * cell_size
    return cell_number, picker_x, picker_y




#!/usr/bin/env python3
"""
    Copyright (C) Ηλίας Ηλιάδης, 2019-09-24; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    This file is part of «Sudoku as puzzle».

    «Sudoku as puzzle» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «Sudoku as puzzle» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «Sudoku as puzzle».  If not, see <http://www.gnu.org/licenses/>.
"""

"""
        WARNING !!!

Since we are in python, any method here will override base class.
So do not declare here methods that exist also in the base class.
"""
#FIXME: correct the version
__version__ = '0.1.28'
VERSIONSTR = 'v. {}'.format(__version__)

try:
    import os
    import sys

    # Gtk and related
    from gi import require_version as gi_require_version
    gi_require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from gi.repository import Gdk, GdkPixbuf, Pango
    import cairo
    gi_require_version('PangoCairo', '1.0')
    from gi.repository import PangoCairo

    # Configuration and message boxes
    from auxiliary import SectionConfig, OptionConfig, MessageBox
    from constants import *

except ImportError as eximp:
    print(eximp)
    sys.exit(-1)

game_cells = []
for acounter in range(81):
    s = TMPDEFAULTPUZZLE[acounter]
    is_constant = True if s != "." else False
    theint = int(s) if s != "." else 0
    game_cells.append ([is_constant, theint, False])
game_cells.insert(0, 1)

def create_board_pixbuf_for_options(pb):
    if pb == None: return

    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, OPTIONS_BOARD_SIZE, OPTIONS_BOARD_SIZE)
    cr = cairo.Context(surface)
    x0 = int(OPTIONS_BOARD_SIZE // 3)
    # x3 = int(x0 // 3)
    for column in range(3):
        for row in range(3):
            cell_number = 1 + row * 3 + column
            apixbuf = pb.picker_pixbufs[cell_number]

            cr.save()
            if cell_number == 1:
                cr.set_source_rgba(*pb.newcolors["RGBA_PICKER_BG"])
            elif (cell_number == 2) or (cell_number == 6):
                cr.set_source_rgba(*pb.newcolors["RGBA_BG"])
            elif (cell_number == 3 ) or (cell_number == 7):
                cr.set_source_rgba(*pb.newcolors["RGBA_BG_SELECTED"])
            elif (cell_number == 4 ) or (cell_number == 8):
                cr.set_source_rgba(*pb.newcolors["RGBA_BG_CONSTANT"])
            else:
                cr.set_source_rgba(*pb.newcolors["RGBA_BG_CONSTANT_SELECTED"])
            if (cell_number>5):
                apixbuf = pb.fitted_pixbufs_R[cell_number]
            else:
                apixbuf = pb.fitted_pixbufs[cell_number]

            # first a square with selected color
            cr.rectangle(x0*column,x0*row, x0, x0)
            cr.fill()
            #then print a scaled num (red or black)
            Gdk.cairo_set_source_pixbuf(cr, apixbuf,x0*column,x0*row)
            cr.paint()
            cr.stroke()
            cr.restore()

    cr.set_source_rgba(0, 0, 0, 1)
    for acounter in range(2):
        cr.save()
        cr.set_line_width(1)
        x1 = x0 * (acounter + 1)
        cr.move_to(x1, 0)
        cr.line_to(x1, OPTIONS_BOARD_SIZE)
        cr.stroke()
        cr.move_to(0, x1)
        cr.line_to(OPTIONS_BOARD_SIZE, x1)
        cr.stroke()
        cr.restore()


    w = h = OPTIONS_BOARD_SIZE #- 4
    cr.set_line_width(3)

    cr.rectangle(2, 2, w-4, h-4)
    cr.stroke()
    board_pixbuf = Gdk.pixbuf_get_from_surface(surface, 0, 0, OPTIONS_BOARD_SIZE, OPTIONS_BOARD_SIZE)
    return board_pixbuf



#!/usr/bin/env python3
"""
    Copyright (C) Ηλίας Ηλιάδης, 2019-09-24; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    This file is part of «Sudoku as puzzle».

    «Sudoku as puzzle» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «Sudoku as puzzle» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «Sudoku as puzzle».  If not, see <http://www.gnu.org/licenses/>.
"""

#FIXME: correct the version
APPID = "org.kekbay.sudokuaspuzzleocp"
APPNAME = "Sudoku as puzzle"
__version__ = '0.1.28'

ERROR_IMPORT_LIBRARIES_FAIL = -1

try:
    import os
    import sys

    # Gtk and related
    from gi import require_version as gi_require_version
    gi_require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from gi.repository import Gdk, GdkPixbuf, GObject, Gio, GLib

    # Localization
    import locale
    import gettext

    #from locale import gettext

    # Configuration and message boxes
    from auxiliary import *
    from constants import *

    #At least a starting window must exist
    from windowmain import WindowMainClass

except ImportError as eximp:
    print("eximp1", eximp)
    sys.exit(ERROR_IMPORT_LIBRARIES_FAIL)

WHERE_AM_I = os.path.dirname(os.path.abspath(os.path.realpath(__file__)))

settings = None # Will keep window related options
options = None # Will keep application wide options in a 'General Options' section

class CustomAppArgs:
    pass

class MyApplication(Gtk.Application):
    #FIXME: fix the docstring.
    """ Main entry of the application. """

    def __init__(self, *args, **kwargs):
        self.custom_args = CustomAppArgs()
        self.custom_args.START_DIR = kwargs['START_DIR']
        self.custom_args.APP_DIR = WHERE_AM_I
        self.id = APPID
        self.custom_args.version = __version__
        self.custom_args.windowmain = None
        self.init_locale()
        self.custom_args.name = APPNAME
        self.custom_args.localizedname = _("Sudoku as puzzle")

        Gtk.Application.__init__(self,
                application_id=self.id,
                flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE)

        # Load the icon file.
        icon_file = os.path.join(self.custom_args.APP_DIR, "icons", "logo.png")
        self.icon = None
        if os.path.exists(icon_file):
            try:
                self.icon = GdkPixbuf.Pixbuf.new_from_file(icon_file)
            except Exception:
                pass

        # Init the modules for settings (globally).
        self.set_config_file()

        self.connect("activate", self.my_activate)

    def init_locale(self):
        LOCALE_DIR = os.path.join(self.custom_args.APP_DIR, 'localedir')
        locale.setlocale(locale.LC_ALL, '')
        locale.bindtextdomain(self.id , LOCALE_DIR)
        #bind locale application wide
        gettext.install(self.id, LOCALE_DIR)

    def set_config_file(self):
        global settings
        settings = SectionConfig(self.id , self.__class__.__name__)
        global options
        options = OptionConfig(self.id)

    def my_activate(self, *args, **kwargs):
        # We only allow a single window and raise any existing ones
        if not self.custom_args.windowmain:
            self.open_windowmain()
        else:
            #probably unnecessary since it prints out in the first open
            #not in the second... so a second open will simply aborts.
            #logging is disk space consuming, so no log provided.
            print("WARNING! Application main window already exists")
        self.custom_args.windowmain.present()

    #FIXME: remove if not needed
    def custom_return_function(self, *args, **kwargs):
        """ Custom function to be triggered before opened window exit.

        Opened "startwindow" can use it to:
        - Return to this class a value.
        - Get from this class a processed value before exit.

        Notes
        -----
        This function will run BEFORE the opened window is destroyed.
        If you want to do something AFTER the window is destroyed,
        you should find other way or trigger.
        """
        # dummy printout
        # print("Game exited")
        # dummy return of my class name
        return self.__class__.__name__

    def do_command_line(self, command_line):
        """ Parse command line and activate the application. """
        start_options = command_line.get_options_dict()
        if start_options.contains("test"):
            # This is printed on the main instance
            # dummy printout
            print("Test argument recieved")
        self.activate()
        return 0

    def open_windowmain(self):
        #create args to send
        some_custom_args = {}
        some_custom_args['a text arg'] = "some text to pass as arg"

        # provide a local function to be triggered on exit
        some_custom_args['trigger_before_exit'] = self.trigger_from_windowmain
        #import the class
        #this is the app file. one window imported already

        #create an instance and add it to self.Application.MyArgs
        #as property
        self.custom_args.windowmain = WindowMainClass(application = self,
            custom_args = some_custom_args)
        self.custom_args.windowmain.mywindow.connect("destroy", self.windowmain_destroyed)
        #open the window
        self.custom_args.windowmain.present()

    def windowmain_destroyed(self, *args, **kwargs):
        """ Custom function to be triggered after opened window is destroyed.

        """
        #self.exit_requested()
        #nullify window as property of self.Application.MyArgs
        self.custom_args.windowmain = None

    def trigger_from_windowmain(self, *args, **kwargs):
        """ Custom function to be triggered before opened window exit.

        Opened "windowmain" can use it to:
        - Return to this class a value.
        - Get from this class a processed value before exit.

        Notes
        -----
        This function will run BEFORE the opened window is destroyed.
        If you want to do something AFTER the window is destroyed,
        you must use windowmain_destroyed.
        """

        # dummy return of my class name
        return self.__class__.__name__



#!/usr/bin/env python3
"""
    Copyright (C) Ηλίας Ηλιάδης, 2019-09-24; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    This file is part of «Sudoku as puzzle».

    «Sudoku as puzzle» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «Sudoku as puzzle» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «Sudoku as puzzle».  If not, see <http://www.gnu.org/licenses/>.
"""

try:
    import os
    import sys
    import math
    from copy import deepcopy

    # Gtk and related
    from gi import require_version as gi_require_version
    gi_require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from gi.repository import Gdk, GdkPixbuf, Pango
    import cairo
    gi_require_version('PangoCairo', '1.0')
    from gi.repository import PangoCairo

    from constants import *

except ImportError as eximp:
    print(eximp)
    sys.exit(-1)


class GamePixBufs:
    def __init__(self, window, stringslist, fontstring, newcolors, font_ratio = 1):
        self.fontstring = fontstring
        self.fix_font_size()
        self.font_ratio = font_ratio
        self.newcolors = newcolors

        self.fgcolor_cairo = self.newcolors["RGBA_FG"]
        self.fgRcolor_cairo = self.newcolors["RGBA_FG_RED"]

        #create a dummy cairo context da
        cr = window.get_window().cairo_create()
        self.bigpixbufs = []
        self.bigpixbufs_R = []
        # self.bigpixbufs_P = []

        self.picker_pixbufs = []

        wh = 0
        pixbufs1 = []
        pixbufs_R = []
        for aletter in stringslist[1:]:
            try:
                apixbuf = self.create_digit(aletter)
                apixbuf_R = self.create_digit(aletter, True)
            except Exception as e:
                print("apixbuf Exception", e)
            wh = max(wh, apixbuf.get_width(), apixbuf.get_height())
            pixbufs1.append(apixbuf)
            pixbufs_R.append(apixbuf_R)
        #insert an empty pixbuf
        theempty = self.anempty_big(pixbufs1[0], wh)
        pixbufs1.insert(0, theempty)
        pixbufs_R.insert(0, theempty)
        for apixbuf in pixbufs1:
            self.bigpixbufs.append(self.resize_bg(apixbuf, wh))
            # self.bigpixbufs_P.append(self.resize_bg(apixbuf, wh, for_picker = True))

        for apixbuf in pixbufs_R:
            self.bigpixbufs_R.append(self.resize_bg(apixbuf, wh))

        # TODO try to set a minimu picker pibuf

    def fix_font_size(self):
        self.fontstring = self.fontstring.split(" ",-1)[0] + " 300"

    def fit_pixbufs(self, ameasure):
        self.fitted_pixbufs = []
        self.fitted_pixbufs_R = []
        self.picker_pixbufs = []
        one_third = int(ameasure / 3) - 2
        for apixbuf in self.bigpixbufs:
            scaled = apixbuf.scale_simple(ameasure,
                    ameasure,
                    GdkPixbuf.InterpType.BILINEAR)
            self.fitted_pixbufs.append(scaled)
        # for apixbuf in self.bigpixbufs_P:
            picker_pixbuf = apixbuf.scale_simple(one_third,
                    one_third,
                    GdkPixbuf.InterpType.BILINEAR)
            self.picker_pixbufs.append(picker_pixbuf)
        for apixbuf in self.bigpixbufs_R:
            scaled = apixbuf.scale_simple(ameasure,
                    ameasure,
                    GdkPixbuf.InterpType.BILINEAR)
            self.fitted_pixbufs_R.append(scaled)

    def create_digit(self, letter, red = False, white = False):
        #pixbuf will be unscaled
        #just a big pixbuf
        fontname = self.fontstring
        left = 20
        top = 0
        #create a temporary pixbuf
        #to find the extends
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 800, 600)
        cr = cairo.Context(surface)

        layout = PangoCairo.create_layout(cr)
        thefontdescr = Pango.font_description_from_string(fontname)

        layout.set_font_description(Pango.font_description_from_string(fontname))
        crcr = layout.get_context()

        amap = crcr.get_font_map()
        afont = amap.load_font (crcr, thefontdescr)

        cr.move_to(left,top)

        if red:
            cr.set_source_rgba(*self.fgRcolor_cairo)
        else:
            cr.set_source_rgba(*self.fgcolor_cairo)
        #avoid problems with left bearing going out of surface
        layout.set_text(" " + letter, -1)

        PangoCairo.update_layout(cr, layout)

        ink_pixel_extends, logical_pixel_extends = layout.get_pixel_extents ()
        PangoCairo.show_layout(cr, layout)

        surface.flush()

        surface2  = cairo.ImageSurface(cairo.FORMAT_ARGB32,
                ink_pixel_extends.width,
                logical_pixel_extends.height)
        cr2 = cairo.Context(surface2)
        cr2.set_source_surface(surface, -(left + ink_pixel_extends.x), -top)
        cr2.rectangle(0, 0, ink_pixel_extends.width, logical_pixel_extends.height)
        cr2.fill()

        digit_pixbuf = Gdk.pixbuf_get_from_surface (surface2,
                0,
                0,
                ink_pixel_extends.width,
                logical_pixel_extends.height)

        return digit_pixbuf

    def resize_bg(self, pixbuf, full_W, for_picker = False):
        font_ratio = self.font_ratio
        if for_picker:
            font_ratio = 1

        old_W = pixbuf.get_width() * font_ratio
        old_H = pixbuf.get_height() * font_ratio
        diff_W = full_W - old_W
        diff_H = full_W - old_H
        dest_x = diff_W // 2
        dest_y = diff_H // 2
        colorspace = pixbuf.get_colorspace()
        bitspersample = pixbuf.get_bits_per_sample()
        new_pixbuf = GdkPixbuf.Pixbuf.new(colorspace, True, bitspersample, full_W, full_W)
        # new_pixbuf.fill(self.all_colors["TRANS_BG_ΙΝΤ"])
        new_pixbuf.fill(rgba_int(self.newcolors["RGBA_BG_TRANSP"]))
        pixbuf.composite(new_pixbuf,dest_x,dest_y,old_W,old_H,dest_x,dest_y,font_ratio,font_ratio,GdkPixbuf.InterpType.BILINEAR , 255)
        return new_pixbuf

    def anempty_big(self, otherpixbuf, wh):
        colorspace = otherpixbuf.get_colorspace()
        bitspersample = otherpixbuf.get_bits_per_sample()
        new_pixbuf = GdkPixbuf.Pixbuf.new(colorspace, True, bitspersample, wh, wh)
        new_pixbuf.fill(rgba_int(self.newcolors["RGBA_BG"]))
        return new_pixbuf

    def resize_letters(self, one_size):
        pass



#!/usr/bin/env python3
"""
    Copyright (C) Ηλίας Ηλιάδης, 2019-09-24; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    This file is part of «Sudoku as puzzle».

    «Sudoku as puzzle» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «Sudoku as puzzle» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «Sudoku as puzzle».  If not, see <http://www.gnu.org/licenses/>.
"""

"""
        WARNING !!!

Since we are in python, any subclass of this class
will take over any method with the same name.
So do not declare here methods that exist also in the subclass.
(subclass is the class that uses the class here as Base.
Imagine the subclass as a super-duper class,
although in computing superclass is the base class)
"""
#FIXME: correct the version
__version__ = '0.1.28'
VERSIONSTR = 'v. {}'.format(__version__)

try:
    import os
    import sys

    # Gtk and related
    from gi import require_version as gi_require_version
    gi_require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from gi.repository import Gdk
    from gi.repository import GObject
    # space for extra Gtk imports

    # Configuration and message boxes
    from auxiliary import *
    from constants import *

    # Load static methods
    from windowmain_statics import *

except ImportError as eximp:
    print(eximp)
    sys.exit(-1)

class WindowMainBase(object):
    #FIXME: fix the docstring.
    """ Main window with all components. """

    def __init__(self, *args, **kwargs):
        # Set the app
        self.myparent = None
        self.custom_args = kwargs['custom_args']

        # bind settings,options to a class variable
        global settings
        settings = self.settings
        global options
        options = self.options

        # Before builder.
        self._run_before_builder()

        # Read GUI from file and retrieve objects from Gtk.Builder
        thebuilder = Gtk.Builder()
        thebuilder.set_translation_domain(self.Application.id)
        try:
            thebuilder.add_from_file(os.path.join(self.Application.custom_args.APP_DIR,
                'ui',
                'windowmain.glade')
                )
        except GObject.GError:
            print("Error reading GUI file")
            raise

        # Fire up the main window
        #We must get the object from builder to set the Application
        #before loading the objects, so that the rest will be loaded
        #using locale
        self.WindowMain = thebuilder.get_object("WindowMain")
        self.WindowMain.set_application(self.Application)
        self._get_from_builder(thebuilder)
        self.mywindow = self.WindowMain
        self._post_initialisations()
        self.mywindow.show()

#********* Auto created "class defs" START ************************************************************
    def _run_before_builder(self):

        self.return_position  = None
        pass

    def _get_from_builder(self, builder):
        """ Create self names for easy access. """
        self.BoxForOptionsInfo = builder.get_object('BoxForOptionsInfo')
        self.BoxMain = builder.get_object('BoxMain')
        self.BoxOnBottom = builder.get_object('BoxOnBottom')
        self.BoxOnTop = builder.get_object('BoxOnTop')
        self.ButtonAbout = builder.get_object('ButtonAbout')
        self.ButtonContinue = builder.get_object('ButtonContinue')
        self.ButtonExit = builder.get_object('ButtonExit')
        self.ButtonNew = builder.get_object('ButtonNew')
        self.ButtonOptions = builder.get_object('ButtonOptions')
        self.GridForCurrentOptions = builder.get_object('GridForCurrentOptions')
        self.LabelDifficulty = builder.get_object('LabelDifficulty')
        self.LabelDummyDifficulty = builder.get_object('LabelDummyDifficulty')
        self.LabelDummyNumbers = builder.get_object('LabelDummyNumbers')
        self.LabelNumbers = builder.get_object('LabelNumbers')
        self.LabelOptionsInfo = builder.get_object('LabelOptionsInfo')
        self.LabelVersion = builder.get_object('LabelVersion')
        self.WindowMain = builder.get_object('WindowMain')

        # Connect signals existing in the Glade file.
        builder.connect_signals(self)

        # Connect generated by OCPgenerator signals:
        # to builder's main window
        # first are the defaults for window
        self.WindowMain.connect('delete-event', self.on_WindowMain_delete_event)
        self.WindowMain.connect('destroy', self.on_WindowMain_destroy)
        self.WindowMain.connect('size-allocate', self.on_WindowMain_size_allocate)
        self.WindowMain.connect('window-state-event', self.on_WindowMain_window_state_event)
        self.ButtonAbout.connect('clicked', self.on_ButtonAbout_clicked)
        self.ButtonContinue.connect('clicked', self.on_ButtonContinue_clicked)
        self.ButtonExit.connect('clicked', self.on_ButtonExit_clicked)
        self.ButtonNew.connect('clicked', self.on_ButtonNew_clicked)
        self.ButtonOptions.connect('clicked', self.on_ButtonOptions_clicked)

    def _post_initialisations(self):
        """ Do some extra initializations.

        Display the version if a labelVersion is found.
        Set defaults (try to load them from a configuration file):
            - Window size and state (width, height and if maximized)
        Load any custom settings from a configuration file.
        """
        if 'parent' in self.custom_args:#Is a child window, get the parent window
            self.myparent = self.custom_args['parent']
            #if has parent check for transient
            # modality can be false, and parent may not be present
            if 'transient' in self.custom_args:
                self.mywindow.set_transient_for(self.myparent)
                if 'modal' in self.custom_args:
                    self.mywindow.set_modal(True)
                    # print("is modal")

        if 'trigger_before_exit' in self.custom_args:
            # must be a function on calling class
            self.trigger_before_exit = self.custom_args['trigger_before_exit']
            self.return_parameters = None

        # Bind message boxes.
        self.MessageBox = MessageBox(self.WindowMain, self.Application)
        self.msg = self.MessageBox.Message
        self.are_you_sure = self.MessageBox.are_you_sure

        # Reset MainWindow to a default or previous size and state.
        width = settings.get('width', 350)
        height = settings.get('height', 350)
        self.WindowMain.set_title(self.Application.custom_args.localizedname)
        self.WindowMain.resize(width, height)
        self.WindowMain.set_icon(self.Application.icon)
        if settings.get_bool('maximized', False):
            self.WindowMain.maximize()
        # Set the label for LabelVersion
        self.LabelVersion.set_label("v." + self.Application.custom_args.version)
        self.LabelVersion.set_tooltip_text(_("Version of this window:") + "\n" + VERSIONSTR)
        # Load any other settings here.

#********* Auto created handlers START *********************************
    def on_ButtonAbout_clicked(self, widget, *args):
        """ Handler for ButtonAbout.clicked. """
        self.MessageBox.AboutBox()
        return True

    def on_ButtonContinue_clicked(self, widget, *args):
        """ Handler for ButtonContinue.clicked. """
        self.open_windowpuzzle(True)
        return True

    def on_ButtonExit_clicked(self, widget, *args):
        """ Handler for ButtonExit.clicked. """
        self.exit_requested()
        return True

    def on_ButtonNew_clicked(self, widget, *args):
        """ Handler for ButtonNew.clicked. """
        self.open_windowpuzzle(False)
        return True

    def on_ButtonOptions_clicked(self, widget, *args):
        """ Handler for ButtonOptions.clicked. """
        self.open_windowoptions()
        return True

    def on_WindowMain_key_release_event(self, widget, event, *args):
        """ Handler for WindowMain.key-release-event. """
        txt = Gdk.keyval_name(event.keyval)
        if type(txt) == type(None):
            return False

        unichar = chr(Gdk.keyval_to_unicode(event.keyval))
        txt = txt.replace('KP_', '')
        if event.get_state() & Gdk.ModifierType.CONTROL_MASK:
            if txt.lower() == "n":
                self.open_windowpuzzle(False)
                return True
            elif txt.lower() == "c":
                if self.ButtonContinue.get_sensitive():
                    self.open_windowpuzzle(True)
                return True
            elif txt.lower() == "p":
                self.open_windowoptions()
                return True
            elif txt.lower() == "q":
                self.exit_requested()
                return True
            elif txt.lower() == "a":
                self.MessageBox.AboutBox()
                return True
#********* Auto created handlers END ***********************************

#********* Standard handlers START *************************************
    def msg_not_yet(self):
        self.msg(_('Not yet implemented'))

    def on_WindowMain_delete_event(self, widget, event, *args):
        """ Handler for our main window: delete-event. """
        return (self.exit_requested())

    def on_WindowMain_destroy(self, widget, *args):
        """ Handler for our main window: destroy. """
        return (self.exit_requested('from_destroy'))

    def on_WindowMain_size_allocate(self, widget, allocation, *args):
        """ Handler for our main window: size-allocate. """
        self.save_my_size()

    def on_WindowMain_window_state_event(self, widget, event, *args):
        """ Handler for our main window: window-state-event. """
        settings.set('maximized',
            ((int(event.new_window_state) & Gdk.WindowState.ICONIFIED) != Gdk.WindowState.ICONIFIED) and
            ((int(event.new_window_state) & Gdk.WindowState.MAXIMIZED) == Gdk.WindowState.MAXIMIZED)
            )
        self.save_my_size()

#********* Standard handlers END ***************************************
#********* Standard exit defs START *********************************************************
    def exit_requested(self, *args, **kwargs):
        """ Final work before exit. """
        self.WindowMain.set_transient_for()
        self.WindowMain.set_modal(False)
        self.set_unhandled_settings()# also saves all settings
        if 'from_destroy' in args:
            return True
        else:
            # Check if we should provide info to caller
            if 'trigger_before_exit' in self.custom_args:
                self.trigger_before_exit(exiting = True,
                    return_parameters = self.return_parameters)
            self.WindowMain.destroy()

    def present(self):
        """ Show the window. """
        pass

    def save_my_size(self):
        """ Save the window size into settings, if not maximized. """
        if not settings.get_bool('maximized', False):
            width, height = self.WindowMain.get_size()
            settings.set('width', width)
            settings.set('height', height)

    def set_unhandled_settings(self):
        """ Set, before exit, any settings not applied during the session.

        Additionally, flush all settings to .conf file.
        """
        # Set any custom settings
        # which where not setted (ex. on some widget's state changed)

        # Save all settings
        settings.save()
#********* Standard exit defs END **************************************
#********* Auto created "class defs" END **************************************************************

    def show_defaults(self):
        DIFFICULTY = 'easy'
        use_dict = "standard"
        DIFFICULTY = options.get("DIFFICULTY", DIFFICULTY)
        if DIFFICULTY not in ["simple", "easy", "intermediate", "expert", "any"]:
            DIFFICULTY = 'easy'

        strings = options.get("custom_dict","")
        strings_list = [x.strip() for x in strings.split(",")]
        if len(strings_list) == 9:
            strings_list.insert(0, ' ')
            THEDICTS['custom'] = strings_list
        use_dict = options.get("use_dict", use_dict)
        self.LabelDifficulty.set_label(_(DIFFICULTY))
        self.LabelNumbers.set_label(','.join(THEDICTS[use_dict][1:]))

        has_good_history = True
        try:
            last_puzzle_string = options.get("last_puzzle_string", "")
            if len(last_puzzle_string.strip()) != 81:
                has_good_history = False
            else:
                last_history_string = options.get("last_history","")
                thelist1 = last_history_string.strip().split(",")
                for amove in thelist1:
                    if len(amove.split(":")) != 4:
                        has_good_history = False
                        break
                if has_good_history:
                    last_history_position = options.get("last_history_position","-1")
                    if int(last_history_position) < 0:
                        has_good_history = False
                    last_history_position = int(last_history_position)
                    if has_good_history:
                        last_difficulty = options.get("last_difficulty", "")
                        if last_difficulty not in ["simple", "easy", "intermediate", "expert"]:
                            has_good_history = False
                        if has_good_history:
                            last_seconds_passed = int(options.get("last_seconds_passed","0"))
                            if last_seconds_passed == 0:
                                has_good_history = False
        except:
            print(_("WARNING! Wrong values for history in configuration file"))
            has_good_history = False
        if has_good_history:
            self.last_game = {'puzzle_string':last_puzzle_string,
                    'history_string':last_history_string,
                    'history_position': last_history_position,
                    'last_seconds_passed': last_seconds_passed,
                    'difficulty':last_difficulty}
        self.ButtonContinue.set_sensitive(has_good_history and (not options.get_bool('solved', True)))

    #call this from wherever event you like
    def open_windowpuzzle(self, continue_last=False):
        #save current position
        self.return_position = self.mywindow.get_position()
        #create args to send
        some_custom_args = {}
        # provide a local function to be triggered on exit
        some_custom_args['trigger_before_exit'] = self.trigger_from_windowpuzzle
        some_custom_args['parent'] = self.mywindow
        if continue_last:
            some_custom_args['continue'] = continue_last
            some_custom_args['last_game'] = self.last_game.copy()

        #import the class
        from windowpuzzle import WindowPuzzleClass
        #create an instance and add it to self.Application.MyArgs
        #as property
        self.Application.custom_args.windowpuzzle = WindowPuzzleClass(application = self.Application,
            custom_args = some_custom_args)
        self.Application.custom_args.windowpuzzle.mywindow.connect("destroy", self.windowpuzzle_destroyed)
        #open the window
        self.Application.custom_args.windowpuzzle.present()
        #hide this window or comment next line if not needed
        self.mywindow.hide()

    def windowpuzzle_destroyed(self, *args, **kwargs):
        """ Custom function to be triggered after opened window is destroyed.

        """
        self.Application.custom_args.windowpuzzle = None
        self.refresh_me()

    def trigger_from_windowpuzzle(self, *args, **kwargs):
        """ Custom function to be triggered before opened window exit.

        Opened "windowpuzzle" can use it to:
        - Return to this class a value.
        - Get from this class a processed value before exit.

        Notes
        -----
        This function will run BEFORE the opened window is destroyed.
        If you want to do something AFTER the window is destroyed,
        you must use windowpuzzle_destroyed.
        """
        # dummy return of my class name
        return self.__class__.__name__

    def refresh_me(self):
        #TODO: create and refresh also results?

        #TODO: return to a position depending on opened child window
        #not in the previous position
        #Return to previous position
        #Cause in case of multiple monitors
        #it returns to the center of the screen
        #in which a "fast mouse move" goes
        if self.return_position:
            self.mywindow.move(*self.return_position)
            self.return_position = None
        self.mywindow.show()
        self.show_defaults()

    #call this from wherever event you like
    def open_windowoptions(self):
        #save current position
        self.return_position = self.mywindow.get_position()
        #create args to send
        some_custom_args = {}
        # provide a local function to be triggered on exit
        some_custom_args['trigger_before_exit'] = self.trigger_from_windowoptions
        #import the class
        from windowoptions import WindowOptionsClass
        #create an instance and add it to self.Application.MyArgs
        #as property
        self.Application.custom_args.windowoptions = WindowOptionsClass(application = self.Application,
            custom_args = some_custom_args)
        self.Application.custom_args.windowoptions.mywindow.connect("destroy", self.windowoptions_destroyed)
        #open the window
        self.Application.custom_args.windowoptions.present()
        #hide this window or comment next line if not needed
        self.mywindow.hide()

    def windowoptions_destroyed(self, *args, **kwargs):
        """ Custom function to be triggered after opened window is destroyed.

        """
        self.Application.custom_args.windowoptions = None
        self.refresh_me()

    def trigger_from_windowoptions(self, *args, **kwargs):
        """ Custom function to be triggered before opened window exit.

        Opened "windowoptions" can use it to:
        - Return to this class a value.
        - Get from this class a processed value before exit.

        Notes
        -----
        This function will run BEFORE the opened window is destroyed.
        If you want to do something AFTER the window is destroyed,
        you must use windowoptions_destroyed.
        """
        # dummy return of my class name
        return self.__class__.__name__

#********* Window class  END***************************************************************************



#!/usr/bin/env python3
from setuptools import setup
import os
import sys

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='sudokuaspuzzle',
      version='0.1.28',
      description='Customizable Sudoku',
      long_description=readme(),
      long_description_content_type='text/markdown',
      url='http://gnu.kekbay.gr/sudokuaspuzzle/versions',
        project_urls={
        'Source': 'https://gitlab.com/xoristzatziki/sudokuaspuzzle',
      },
      author='Ηλίας Ηλιάδης',
      author_email='iliadis@kekbay.gr',
      classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: X11 Applications :: GTK',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.6',
        'Topic :: Games/Entertainment :: Puzzle Games',
      ],
      license='GPLv3',
      packages=['sudokuaspuzzle'],
      include_package_data=True,
      entry_points = {
        'console_scripts': [
            'sudokuaspuzzleC = sudokuaspuzzle.entrypoint:main'],
        'gui_scripts': [
            'sudokuaspuzzle = sudokuaspuzzle.entrypoint:main',
        ]
      },
      zip_safe=False)



#!/usr/bin/env python3
"""
    Copyright (C) Ηλίας Ηλιάδης, 2019-09-24; Ηλίας Ηλιάδης <iliadis@kekbay.gr>

    This file is part of «Sudoku as puzzle».

    «Sudoku as puzzle» is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    «Sudoku as puzzle» is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with «Sudoku as puzzle».  If not, see <http://www.gnu.org/licenses/>.
"""

"""
        WARNING !!!

Since we are in python, any method here will override base class.
So do not declare here methods that exist also in the base class.
"""
#FIXME: correct the version
__version__ = '0.1.28'
VERSIONSTR = 'v. {}'.format(__version__)

try:
    import os
    import sys

    # Configuration and message boxes
    from auxiliary import *
    from constants import *

except ImportError as eximp:
    print(eximp)
    sys.exit(-1)

